(* ::Package:: *)

(*:Date: Sun May 17 16:47:16 MET DST 1998 *)
(*:Version 1.1: May 24, 2012 *)
(*This is only the part " Parametrically Defined Space Curves" , "Parametrically Defined Curves in R^n",
and "Approximations" of Alfred Gray's package CURVES.m. *)

(* Title: Curves3D.m *)

(*:Author: Alfred Gray *)

(* Copyright 1994-1998 by Alfred Gray *)

(*:Package Version: 1.0 *)

(*:Mathematica Versions: 2.2, 3.0 *)


(* Package Version 1.1. 
Adapted to Mathematica Version 7.0  
by Rolf Sulanke, starting May 24, 2012 *)

(* August 12, 2016. Alfred Gray's original package  CURVES.m divided into two packages:
Curves2D.m, the plane curves,
Curves3D.m, the other object, mostly space curves. Rolf Sulanke. *)


(* Changes in V. 1.1: No additional packages loaded. The Print command atthe beginning is 
commented out now, instead use `?CURVES* in the notebooks. In veriflat::usage the word "same" added. 
The command Protect["`*"]; added before the END of the packages protecting all the symbols defined 
in the context of the package.*)

(*:Summary: 
This package consists of definitions of parametrically defined space curves.
*)

(*:Keywords:
agnesi, agnesiimplicit, airyspiral, alysoid, alysoidnds, alysoidprime, archimedesspiral, archimedesspiralpolar, astroid, astroidimplicit, ast3d, baseballseam, bellcurve, besselcurve, besseln, bernstein, beziercurve, bicorn, bicylinder, biquadratic, bow, bowpolar, bowtie, bulletnose, bulletnoseimplicit, cardioid, cardioidimplicit, cardioidpolar, cardioidunitspeed, cartesianimplicit, cartesianoval, cassini, cassiniimplicit, cassinipolar, catenary, catenaryunitspeed, cayleysextic, circle, circleimplicit, circleinvolute, circleinvoluteunitspeed, circleunitspeed, cissoid, cissoidimplicit, clelia, clothoid, clothoidnds, clothoidprime, cnccoprofile, cnchyprofile, cochleoid, cochleoidpolar, conchoid, conchoidimplicit, conicalhelix, coshspiral, cosn, cothspiral, cpcprofile, crosscurve, crosscurveimplicit, cubicalellipse, cubicparabola, cubicparabolaimplicit, curves3d, cycloid, cycloidunitspeed, delaunay, delaunaykappa2, delaunaynds, deltoid, deltoidbis, deltoidimplicit, deltoidinvolute, deltoidunitspeed, devilimplicit, diamond, eight, eightknot, elasticainflect, elasticanoninflect, ellipse, ellipseimplicit, ellipseinc, epicycloid, epicycloidunitspeed, epispiral, epispiralpolar, epitrochoid, epitrochoid3d, fermatspiral, fermatspiralpolar, folium, foliumimplicit, foliumpolar, genparabola, genhelix, hankelspiral, helix, helixunitspeed, hippopede, hippopedeimplicit, horopter, hyperbola, hyperbolabis, hyperbolaimplicit, hyperbolicspiral, hyperbolicspiralpolar, hypocycloid, hypocycloidinvolute, hypocycloidunitspeed, hypotrochoid, icurvs2d, kampyle, kampylepolar, keplerimplicit, keplerorbit, keplerorbitpolar, kochsnowflake, lehrkappa2, lehrnds, lemniscate, lemniscatebis, lemniscateimplicit, lemniscatepolar, lemniscateunitspeed, limacon, limaconimplicit, limaconpolar, line, line3d, lissajous, lissajous3d, lituus, lituuspolar, logistic, logspiral, logspiralpolar, logspiralunitspeed, nephroid, nephroidimplicit, nephroidpolar, nephroidunitspeed, newtonphillipsimplicit, ngon, nielsenspiral, oneparametersubgroup, pacman, parabola, parabolaimplicit, parabolicspiral, perseusimplicit, piriform, piriformimplicit, predatorpreyimplicit, pseudocatenary, pseudocatenaryprime, pseudospheregeodesic, pseudosphericalloxodrome, rectangle, regularpolygon, reuleauxpolygon, rose, rosepolar, sc, scarab, scimplicit, seiffertspiral, semicubic, semicubicimplicit, serpentine, serpentineimplicit, sinhspiral, sinn, sinoval, sphericalcardioid, sphericalellipse, sphericalhelix, sphericalloxodrome, sphericalloxodromeunitspeed, sphericalnephroid, sphericalspiral, spring, strophoid, strophoidimplicit, tanhspiral, teardrop, teeth, tennisballseam, tooth, torusknot, tractrix, tractrixminus, tractrixplus, tractrixunitspeed, triangle, trident, tridentimplicit, trisectrix, tschirnhausen, twicubic, twistedn, veryflat, viviani, watt, wattimplcit, wigglyellipse, zcprofile
*)

BeginPackage["Curves3D`"]

(* ,"CSPROGS`","PLTPROGS`","Graphics`Graphics`","Graphics`ImplicitPlot`" has been i the BeginPackage List; deleted by R. 
Sulanke, since obsolete now. *)



(* Parametrically Defined Space Curves *)

    ast3d::usage="t->ast3d[n,a,b][t] is a space curve analog of the plane curve astroid."

    baseballseam::usage="t->baseballseam[a,c][t] is an approximation to the seam on a baseball of radius a."

    besselcurve::usage="t->besselcurve[a,b,c][t] is a space curve whose components are Bessel functions."

    bicylinder::usage="t->bicylinder[a,b,pm][t] is a parametrization of the intersection of the circular cylinders x^2 + y^2 = a^2 and y^2 + z^2 = b^2. One component is obtained by taking pm = +1, the other by taking pm = -1."

    biquadratic::usage="t->biquadratic[a,k][t] is a parametrization of the
intersection of the circular cylinder x^2 + y^2 = a^2 and the elliptic cylinder
x^2 + k^2*z^2 = a^2."

    clelia::usage="t->clelia[a][m,n][t] is a clelia of type (m,n) on a sphere of radius a. The number of petals is n."

    conicalhelix::usage="t->conicalhelix[a,b,c,d][t] is a parametrization of a helix on a cone over t->ellipse[a,b][t]. The slant of the cone is determined by c and the amount of twisting is determined by d."

    cubicalellipse::usage="t->cubicalellipse[a,b,c,d,e,f,pm][t] is a cubical ellipse for pm = 1 and a cubical hyperbola for pm = -1."

    eightknot::usage="t->eightknot[t] is a parametrization of an eight knot."

    epitrochoid3d::usage="t->epitrochoid3d[a,b,h,omega][t] is the parametrized curve traced out by a point P attached to a circle of radius b rolling along a fixed circle of radius a. The rolling circle lies in a plane whose inclination with the plane of the fixed circle is omega, and h is the distance of P to the center of the rolling circle. The curve t->epitrochoid3d[a,b,h,0][t] is a (planar) epitrochoid and the curve epitrochoid3d[a,b,h,Pi][t] is a (planar) hypotrochoid."

    genhelix::usage="t->genhelix[a,b][f][t] is a generalized elliptical helix with slant function f."

    helix::usage="t->helix[a,b][t] is a circular helix of radius a and slant b. t->helix[a,b,c][t] is an elliptical helix of slant c."

    helixunitspeed::usage="s->helixunitspeed[a,b][s] is a unit-speed parametrization of a circular helix of radius a and slant b."

    horopter::usage="t->horopter[a,b][t] is a horopter. It is the intersection of the hyperbolic paraboloid y==b*x*z and the cylinder over the curve y^2==x*(2*a - x)."

    line3d::usage="t->line3d[a1,b1,c1,a2,b2,c2][t] is a parametrized line in R^3 with line3d[a1,b1,c1,a2,b2,c2][0]={a2,b2,c2} and line3d[a1,b1,c1,a2,b2,c2]'[0]={a1,b1,c1}."

    lissajous3d::usage="t->lissajous3d[n,d,a,b,c][t] is a space curve analog of Lissajous curve."

    oneparametersubgroup::usage="t->oneparametersubgroup[mat,vec][t] is the
curve in R^n formed by the one parameter subgroup t->exp(t*mat) acting on a vector v in R^n, where mat is an n by n matrix."

    pseudospheregeodesic::usage="t->pseudospheregeodesic[a,c,u0,pm][t] is a geodesic of slant c on a pseudosphere of constant negative curvature -1/a^2. The parameter pm is plus or minus 1."

    pseudosphericalloxodrome::usage="t->pseudosphericalloxodrome[a,b][t] is a loxodrome on a pseudosphere of constant negative curvature -1/a^2."

    seiffertspiral::usage="t->seiffertspiral[a,k][t] is Seiffert's spiral of slant k on an sphere of radius a."

    sphericalcardioid::usage="t->sphericalcardioid[a][t] is a spherical curve on a sphere of radius 3|a|. It is a curve of constant slope that projects to t->cardioid[-a][t + Pi]."

    sphericalellipse::usage="t->sphericalellipse[a,b,c,pm][t]
is a parametrization of the intersection of the sphere x^2 + y^2 + z^2= a^2 and the elliptical cylinder x^2/b^2 + y^2/c^2 = 1. One component is obtained by taking pm = +1, the other by taking pm = -1."

    sphericalhelix::usage="t->sphericalhelix[a,b][t] is a spherical helix on a sphere of radius |a + 2*b|. It is a curve of constant slope that projects to t->epicycloid[a,b][t]."

    sphericalloxodrome::usage="t->sphericalloxodrome[a,b,c][t] is a loxodrome on a sphere of radius a. It is the image under the stereographic injection {u,v}->stereographicsphere[a][u,v] of the logarithmic spiral t->logspiral[b,c][t]. A sphericalloxodrome meets each meridian of the sphere at the same angle."

    sphericalloxodromeunitspeed::usage="s->sphericalloxodromeunitspeed[a,b,c][s] is a unit-speed parametrization of a loxodrome on a sphere of radius a."

    sphericalnephroid::usage="t->sphericalnephroid[a][t] is a spherical helix on a sphere of radius 4|a|. It is a curve of constant slope that projects to t->nephroid[a][t]."

    sphericalspiral::usage="t->sphericalspiral[a][m,n][t] is a spherical spiral of type (m,n) on a sphere of radius a."

    tennisballseam::usage="t->tennisballseam[n,a,b,c][t] is a curve on the unit sphere in R^3.  It is the normal to t->wigglyellipse[n,a,b,c][t]. For those values of n, a, b and c for which the curve has no self-intersections, the curve (parametrized on 0<=t<=2*Pi) is the boundary of subsets of the unit sphere of equal area."

    torusknot::usage="t->torusknot[a,b,c][p,q][t] is a torus knot of type (p,q) on an elliptical torus of type (a,b,c)."

    twicubic::usage="t->twicubic[t] is the twisted cubic t->{t,t^2,t^3}."

    veryflat::usage="t->veryflat[1][t] is the curve t->{t,0,5*E^(-1/t^2)}.
t->veryflat[2][t] is the curve t->If[t<0,{t,5*E^(-1/t^2),0},
If[t==0,{0,0,0},{t,0,5*E^(-1/t^2)}]]. These curves both have the same curvature and torsion zero, but are not congruent by a Euclidean motion."

    viviani::usage="t->viviani[a][t] is a Viviani curve on a sphere of radius 2*a. It is the intersection of the cylinder (x - a)^2 + y^2==a^2 and the sphere x^2 + y^2 + z^2==4*a^2."

    wigglyellipse::usage="t->wigglyellipse[n,a,b,c][t] is an ellipse perturbed vertically by a sin curve. The normal to t->wigglyellipse[n,a,b,c][t] divides the unit sphere into two regions of equal area."

(* Parametrically Defined Curves in R^n*)

    besseln::usage="t->besseln[n,a][t] is the curve t->a*{BesselJ[0,t],...,BesselJ[n - 1,t]}."

    bernstein::usage="t->bernstein[p,q[t]] is the (p,q)-th Bernstein polynomial."

    beziercurve::usage="t->beziercurve[pts][t] is the Bezier curve in R^n determined by pts, where pts is a list of points in R^n."

    cosn::usage="t->cosn[n,a][t] is the curve t->a*{Cos[t],Cos[2*t],...,Cos[n*t]}."

    beziercurve::usage="t->beziercurve[pts][t] is the Bezier curve in R^n determined by pts, where pts is a list of points in R^n."

    lagrangecurve::usage="t->hermitecurve[p0,m0,m1,p1][t] is the Hermite curve in R^n determined by p0 and p1 with directions m0 and m1."

    sinn::usage="t->sinn[n,a][t] is the curve t->a*{Sin[t],Sin[2*t],...,Sin[n*t]}."

    twistedn::usage="t->twistedn[n,a][t] is the curve t->a*{t,t^2,...,t^n}."

(* Approximations *)

   regularpolygon::usage="regularpolygon[n,a] is a regular plolygon with n vertices circumscribed by a circle of radius a."

   teeth::usage="teeth[m,t][shortlines] is the mth iterate of tooth of height t."

   tooth::usage="tooth[t][Line[{{a,b},{c,d}}]] constructs from the line Line[{{a,b},{c,d}}] a broken line with a tooth of height t in the middle."

   kochsnowflake::usage="kochsnowflake[m] is a Koch snow flake constructed as the mth iterate of an equilateral triangle using a tooth of height (3 + Sqrt[3])/6."


Begin["`Private`"]


(* Parametrically Defined Space Curves *)

    ast3d[n_,a_,b_][t_]:= {a*Cos[t]^n,b*Sin[t]^n,Cos[2*t]}

    ast3d[n_,a_,b_][tmin_,tmax_,opt___]:=
        ParametricPlot3D[Evaluate[ast3d[n,a,b][t]],
            {t,tmin,tmax},opt,Axes->None,PlotPoints->150]

    baseballseam[a_,c_][t_]:=
        {a*Sin[Pi/2 - (Pi/2 - c)*Cos[t]]*Cos[t/2 + c*Sin[2*t]],
         a*Sin[Pi/2 - (Pi/2 - c)*Cos[t]]*Sin[t/2 + c*Sin[2*t]],
         a*Cos[Pi/2 - (Pi/2 - c)*Cos[t]]}

    baseballseam[a_,c_][tmin_,tmax_,opt___]:=
        ParametricPlot3D[Evaluate[baseballseam[a,c][t]],
            {t,tmin,tmax},opt,Axes->None,PlotPoints->150]

    besselcurve[a_,b_,c_][t_]:=
         {BesselJ[a,t],BesselJ[b,t],BesselJ[c,t]}

    besselcurve[a_,b_,c_][tmin_,tmax_,opt___]:=
        ParametricPlot3D[Evaluate[besselcurve[a,b,c][t]],
            {t,tmin,tmax},opt,Axes->None,PlotPoints->150]

    bicylinder[a_,b_,pm_][t_]:= {a*Cos[t],a*Sin[t],
                                 pm*Sqrt[b^2 - a^2*Sin[t]^2]}

    bicylinder[a_,b_,pm_][tmin_,tmax_,opt___]:=
        ParametricPlot3D[Evaluate[bicylinder[a,b,pm][t]],
            {t,tmin,tmax},opt,Axes->None,PlotPoints->150]

    biquadratic[a_,k_][t_]:=
           {a*JacobiSN[t,k],a*JacobiCN[t,k],a*JacobiDN[t,k]}

    biquadratic[a_,k_][tmin_,tmax_,opt___]:=
        ParametricPlot3D[Evaluate[biquadratic[a,k][t]],
            {t,tmin,tmax},opt,Axes->None,PlotPoints->150]

    clelia[a_][m_,n_][t_]:=
           {a*m*Cos[t]*Sin[n*t],a*m*Sin[t]*Sin[n*t],
            a*Sqrt[1 - m^2*Sin[n*t]^2]}

    clelia[a_][m_,n_][tmin_,tmax_,opt___]:=
        ParametricPlot3D[Evaluate[clelia[a][m,n][t]],
            {t,tmin,tmax},opt,Axes->None,PlotPoints->150]

    conicalhelix[a_,b_,c_,d_][t_]:= {a*t*Cos[d*t],b*t*Sin[d*t],c*t}

    conicalhelix[a_,b_,c_,d_][tmin_,tmax_,opt___]:=
        ParametricPlot3D[Evaluate[conicalhelix[a,b,c,d][t]],
            {t,tmin,tmax},opt,Axes->None,PlotPoints->500]

    cubicalellipse[a_,b_,c_,d_,e_,f_,pm_][t_]:=
           {a*t/(1 + pm*t^2),
           (b*t + c*t^2)/(1 + pm*t^2),
           (d*t + e*t^2 + f*t^3)/(1 + pm*t^2)}

    cubicalellipse[a_,b_,c_,d_,e_,f_,pm_][tmin_,tmax_,opt___]:=
        ParametricPlot3D[Evaluate[cubicalellipse[a,b,c,d,e,f,pm][t]],
            {t,tmin,tmax},opt,Axes->None,PlotPoints->150]

    eightknot[t_]:=
        {10*(Cos[t] + Cos[3*t]) + Cos[2*t] + Cos[4*t],
         6*Sin[t] + 10*Sin[3*t],
         4*Sin[3*t]*Sin[5*t/2] + 4*Sin[4*t] - 2*Sin[6*t]}

    eightknot[tmin_,tmax_,opt___]:=
        ParametricPlot3D[Evaluate[eightknot[t]],{t,tmin,tmax},opt,
            Axes->None,PlotPoints->150]

    epitrochoid3d[a_,b_,h_,omega_][t_]:=
        {Cos[t]*(a + b*Cos[omega]) - 
         h*(Cos[t]*Cos[a*t/b]*Cos[omega] - Sin[t]*Sin[a*t/b]),
         Sin[t]*(a + b*Cos[omega]) - 
         h*(Sin[t]*Cos[a*t/b]*Cos[omega] + Cos[t]*Sin[a*t/b]),
        (b - h*Cos[a*t/b])*Sin[omega]}

    epitrochoid3d[a_,b_,h_,omega_][tmin_,tmax_,opt___]:=
        ParametricPlot3D[Evaluate[epitrochoid3d[a,b,h,omega][t]],
            {t,tmin,tmax},opt,Axes->None,PlotPoints->150]

    genhelix[a_,b_][f_][t_]:= {a*Cos[t],b*Sin[t],f[t]}

    helix[a_,c_][t_]:= {a*Cos[t],a*Sin[t],c*t}

    helix[a_,b_,c_][t_]:= {a*Cos[t],b*Sin[t],c*t}

    helix[a_,b_,c_][tmin_,tmax_,opt___]:=
        ParametricPlot3D[Evaluate[helix[a,b,c][t]],
            {t,tmin,tmax},opt,Axes->None,PlotPoints->150]

    helixunitspeed[a_,b_][s_]:=
        {a*Cos[s/Sqrt[a^2 + b^2]],a*Sin[s/Sqrt[a^2 + b^2]],
         b*s/Sqrt[a^2 + b^2]}

    horopter[a_,b_][t_]:=
        {2*a/(1 + b^2*t^2),2*a*b*t/(1 + b^2 t^2),t}

    horopter[a_,b_][tmin_,tmax_,opt___]:=
        ParametricPlot3D[Evaluate[horopter[a,b][t]],
            {t,tmin,tmax},opt,Axes->None,PlotPoints->150]

    line3d[a1_,b1_,c1_,a2_,b2_,c2_][t_]:=
         {a1*t + a2,b1*t + b2,c1*t + c2}

    lissajous3d[n_,d_,a_,b_,c_][t_]:=
         {a*Sin[n*t + d],b*Sin[t],c*Cos[t]} 

    lissajous3d[n_,d_,a_,b_,c_][tmin_,tmax_,opt___]:=
        ParametricPlot3D[Evaluate[lissajous3d[n,d,a,b,c][t]],
            {t,tmin,tmax},opt,Axes->None,PlotPoints->150]

    oneparametersubgroup[mat_,vec_][t_]:=
        ComplexExpand[Normal[MatrixExp[Rationalize[t*mat]]].vec]

    pseudospheregeodesic[a_,c_,u0_,pm_][t_]:=
        {a*Sin[t]*Cos[u0 + pm*Sqrt[a^2 - 2*c^2 - a^2*Cos[2*t]]/
         (Sqrt[2]*c*Sin[t])],
         a*Sin[t]*Sin[u0 + pm*Sqrt[a^2 - 2*c^2 - a^2*Cos[2*t]]/
         (Sqrt[2]*c*Sin[t])],a*(Cos[t] + Log[Tan[t/2]])}

    pseudospheregeodesic[a_,c_,u0_,pm_][tmin_,tmax_,opt___]:=
        ParametricPlot3D[Evaluate[pseudospheregeodesic[a,c,u0,pm][t]],
            {t,tmin,tmax},opt,Axes->None,PlotPoints->150]

    pseudosphericalloxodrome[a_,b_][t_]:=
         {a*Cos[b*Cosh[t]]/Cosh[t],
          a*Sin[b*Cosh[t]]/Cosh[t],a*(t - Tanh[t])}

    pseudosphericalloxodrome[a_,b_][tmin_,tmax_,opt___]:=
        ParametricPlot3D[Evaluate[pseudosphericalloxodrome[a,b][t]],
            {t,tmin,tmax},opt,Axes->None,PlotPoints->500]

    seiffertspiral[a_,k_][t_]:=
        {a*Cos[t]*JacobiSN[t,k],a*Sin[t]*JacobiSN[t,k],
         a*JacobiCN[t,k]}

    seiffertspiral[a_,k_][tmin_,tmax_,opt___]:=
        ParametricPlot3D[Evaluate[seiffertspiral[a,k][t]],
            {t,tmin,tmax},opt,Axes->None,PlotPoints->150]

    sphericalcardioid[a_][t_]:=
        {2*a*Cos[t] - a*Cos[2*t],2*a*Sin[t] - a*Sin[2*t],
         Sqrt[8]*a*Cos[t/2]}

    sphericalcardioid[a_][tmin_,tmax_,opt___]:=
        ParametricPlot3D[Evaluate[sphericalcardioid[a][t]],
            {t,tmin,tmax},opt,Axes->None,PlotPoints->150]

    sphericalellipse[a_,b_,c_,pm_][t_]:=
        {b*JacobiCN[t,(c^2 - b^2)/(a^2 - b^2)],
         c*JacobiSN[t,(c^2 - b^2)/(a^2 - b^2)],
         pm*Sqrt[a^2 - b^2]*JacobiDN[t,(c^2 - b^2)/(a^2 - b^2)]}

    sphericalellipse[a,b,c,pm][tmin_,tmax_,opt___]:=
        ParametricPlot3D[Evaluate[sphericalellipse[a,b,c,pm][t]],
            {t,tmin,tmax},opt,Axes->None,PlotPoints->150]

    sphericalhelix[a_,b_][t_]:=
        {(a + b)*Cos[t] - b*Cos[(a + b)*t/b],
         (a + b)*Sin[t] - b*Sin[(a + b)*t/b],
          2*Sqrt[a*b + b^2]*Cos[a*t/(2*b)]}

    sphericalhelix[a_,b_][tmin_,tmax_,opt___]:=
        ParametricPlot3D[Evaluate[sphericalhelix[a,b][t]],
            {t,tmin,tmax},opt,Axes->None,PlotPoints->150]

    sphericalloxodrome[a_,b_,c_][t_]:=
        {2*a*b*E^(c*t)*Cos[t]/(1 + b^2*E^(2*c*t)),
         2*a*b*E^(c*t)*Sin[t]/(1 + b^2*E^(2*c*t)),
         a*(-1 + b^2*E^(2*c*t))/(1 + b^2*E^(2*c*t))}

    sphericalloxodrome[a_,b_,c_][tmin_,tmax_,opt___]:=
        ParametricPlot3D[Evaluate[sphericalloxodrome[a,b,c][t]],
            {t,tmin,tmax},opt,Axes->None,PlotPoints->150]

    sphericalloxodromeunitspeed[a_,b_,c_][s_]:=
        {2*a*Cos[Log[Tan[c*s/(2*a*Sqrt[1 + c^2])]]/c]*
          Tan[c*s/(2*a*Sqrt[1 + c^2])]/
          (1 + b^2*Tan[c*s/(2*a*Sqrt[1 + c^2])]^2), 
        2*a*Sin[Log[Tan[c*s/(2*a*Sqrt[1 + c^2])]]/c]*
          Tan[c*s/(2*a*Sqrt[1 + c^2])]/
          (1 + b^2*Tan[c*s/(2*a*Sqrt[1 + c^2])]^2), 
          -a*Cos[c*s/(a*Sqrt[1 + c^2])]}

    sphericalnephroid[a_][t_]:=
        {3*a*Cos[t] - a*Cos[3*t],3*a*Sin[t] - a*Sin[3*t],
         Sqrt[12]*a*Cos[t]}

    sphericalnephroid[a_][tmin_,tmax_,opt___]:=
        ParametricPlot3D[Evaluate[sphericalnephroid[a][t]],
            {t,tmin,tmax},opt,Axes->None,PlotPoints->150]

    sphericalspiral[a_][m_,n_][t_]:=
          {a*Cos[m*t]*Cos[n*t],a*Sin[m*t]*Cos[n*t],a*Sin[n*t]}

    sphericalspiral[a_][m_,n_][tmin_,tmax_,opt___]:=
        ParametricPlot3D[Evaluate[sphericalspiral[a][m,n][t]],
            {t,tmin,tmax},opt,Axes->None,PlotPoints->150]

    tennisballseam[n_,a_,b_,c_][t_]:=
          {-a*(Cos[t]*(b^2 + c^2*n^2*Cos[n*t]^2) + 
           c^2*n^3*Sin[t]*Sin[2*n*t]/2)/
           (Sqrt[b^2*Cos[t]^2 + c^2*n^2*Cos[n*t]^2 + 
            a^2*Sin[t]^2]*Sqrt[a^2*b^2*Cos[t]^4 + 
            b^2*c^2*n^2*Cos[n*t]^2*Sin[t]^2 + 
            a^2*b^2*Sin[t]^4 + a^2*c^2*n^4*Sin[t]^2*Sin[n*t]^2 + 
            Cos[t]^2*(a^2*c^2*n^2*Cos[n*t]^2 + 
            b^2*(2*a^2*Sin[t]^2 + c^2*n^4*Sin[n*t]^2)) + 
            a^2*c^2*n^3*Sin[2*t]*Sin[2*n*t]/2 - 
            b^2*c^2*n^3*Sin[2*t]*Sin[2*n*t]/2]),
            b*(-2*(a^2 + c^2*n^2*Cos[n*t]^2)*Sin[t] + 
            c^2*n^3*Cos[t]*Sin[2*n*t])/
            (2*Sqrt[b^2*Cos[t]^2 + c^2*n^2*Cos[n*t]^2 + 
            a^2*Sin[t]^2]*Sqrt[a^2*b^2*Cos[t]^4 + 
            b^2*c^2*n^2*Cos[n*t]^2*Sin[t]^2 + 
            a^2*b^2*Sin[t]^4 + a^2*c^2*n^4*Sin[t]^2*Sin[n*t]^2 + 
            Cos[t]^2*(a^2*c^2*n^2*Cos[n*t]^2 + 
            b^2*(2*a^2*Sin[t]^2 + c^2*n^4*Sin[n*t]^2)) + 
            a^2*c^2*n^3*Sin[2*t]*Sin[2*n*t]/2 - 
            b^2*c^2*n^3*Sin[2*t]*Sin[2*n*t]/2]),
            -c*n*((a^2 - b^2)*Cos[t]*Cos[n*t]*Sin[t] + 
            b^2*n*Cos[t]^2*Sin[n*t] + a^2*n*Sin[t]^2*Sin[n*t])/
            (Sqrt[b^2*Cos[t]^2 + c^2*n^2*Cos[n*t]^2 + 
            a^2*Sin[t]^2]*Sqrt[a^2*b^2*Cos[t]^4 + 
            b^2*c^2*n^2*Cos[n*t]^2*Sin[t]^2 + 
            a^2*b^2*Sin[t]^4 + a^2*c^2*n^4*Sin[t]^2*Sin[n*t]^2 + 
            Cos[t]^2*(a^2*c^2*n^2*Cos[n*t]^2 + 
            b^2*(2*a^2*Sin[t]^2 + c^2*n^4*Sin[n*t]^2)) + 
            a^2*c^2*n^3*Sin[2*t]*Sin[2*n*t]/2 - 
            b^2*c^2*n^3*Sin[2*t]*Sin[2*n*t]/2])}

    tennisballseam[n_,a_,b_,c_][tmin_,tmax_,opt___]:=
        ParametricPlot3D[Evaluate[tennisballseam[n,a,b,c][t]],
            {t,tmin,tmax},opt,Axes->None,PlotPoints->150]

    torusknot[a_,b_,c_][p_,q_][t_]:=
         {(a + b*Cos[q*t])*Cos[p*t],
          (a + b*Cos[q*t])*Sin[p*t],c*Sin[q*t]}

    torusknot[a_,b_,c_][p_,q_][tmin_,tmax_,opt___]:=
        ParametricPlot3D[Evaluate[torusknot[a,b,c][p,q][t]],
            {t,tmin,tmax},opt,Axes->None,PlotPoints->150]

(*
     tprime[a_,b_][v_]:=1/Sqrt[-Cos[v]*(a/b + Cos[v])]


    torusasym[c_,d_,a_,b_][s_]:=
         Module[{uu,tmp1,tmp2,ss},
             tmp1=NDSolve[{uu'[ss]==tprime[a,b][ss],
               uu[c]==d},uu,{ss,0.5001Pi,1.4999Pi},
               MaxSteps->2000];
               tmp2=First[uu[s] /. tmp1];
		{(a + b*Cos[s])*Cos[tmp2],(a + b*Cos[s])*Sin[tmp2],
		b*Sin[s]}]
*)

    twicubic[t_]:= {t,t^2,t^3}

    twicubic[tmin_,tmax_,opt___]:=
        ParametricPlot3D[Evaluate[twicubic[t]],{t,tmin,tmax},opt,
            Axes->None,PlotPoints->150]

    veryflat[1][t_]:= {t,0,5*E^(-1/t^2)}

    veryflat[1][tmin_,tmax_,opt___]:=
        ParametricPlot3D[Evaluate[veryflat[1][t]],
            {t,tmin,tmax},opt,Axes->None,PlotPoints->150]

    veryflat[2][t_]:= If[t<0,{t,5*E^(-1/t^2),0},
                     If[t==0,{0,0,0},{t,0,5*E^(-1/t^2)}]]

    veryflat[2][tmin_,tmax_,opt___]:=
        ParametricPlot3D[Evaluate[veryflat[2][t]],
            {t,tmin,tmax},opt,Axes->None,PlotPoints->150]

    viviani[a_][t_]:= {a*(1 + Cos[t]),a*Sin[t],2*a*Sin[t/2]}

    viviani[a_][tmin_,tmax_,opt___]:=
        ParametricPlot3D[Evaluate[viviani[a][t]],
            {t,tmin,tmax},opt,Axes->None,PlotPoints->150]

    wigglyellipse[n_,a_,b_,c_][t_]:= {a*Cos[t],b*Sin[t],c*Sin[n*t]}

    wigglyellipse[n_,a_,b_,c_][tmin_,tmax_,opt___]:=
        ParametricPlot3D[Evaluate[wigglyellipse[n,a,b,c][t]],
            {t,tmin,tmax},opt,Axes->None,PlotPoints->150]

(* Parametrically Defined Curves in R^n*)

    besseln[n_,a_][t_]:= Table[a*BesselJ[k,t],{k,0,n - 1}]

    bernstein[p_,q_][t_]:=Binomial[p,q]*t^q*(1 - t)^(p - q)
    
    beziercurve[pts_][t_]:= Module[{n=Length[pts]},
         Sum[bernstein[n - 1,k][t]*pts[[k + 1]],
         {k,0,n - 1}]]

         cosn[n_,a_][t_]:= Table[a*Sin[k*t],{k,1,n}]

    hermitecurve[p0_,m0_,m1_,p1_][t_]:=
        (bernstein[3,0][t] + bernstein[3,1][t])*p0 +
        bernstein[3,1][t]*m0 - bernstein[3,2][t]*m1 +
        (bernstein[3,2][t] + bernstein[3,3][t])*p1

    ell[ts_,i_,t_]:=Product[(ts[[j]]-t)/(ts[[j]]-ts[[i]]),{j,1,i-1}]*
        Product[(ts[[j]]-t)/(ts[[j]]-ts[[i]]),{j,i+1,Length[ts]}]
		
    lagrangecurve[pts_,ts_][t_]:= Module[{n=Length[pts]},
         Sum[ell[ts,k+1,t]*pts[[k+1]],{k,0,n - 1}]]

    sinn[n_,a_][t_]:= Table[a*Sin[k*t],{k,1,n}]

    twistedn[n_,a_][t_]:= Table[a*t^k,{k,1,n}]

(* Approximations *)

    regularpolygon[n_,a_]:=
        approxcurve[0,2*Pi,n][circle[a]]

    teeth[m_,t_][shortlines_]:=
	    Nest[Flatten[(tooth[t]/@#)]&,shortlines,m]

(*
    tooth[t_][Line[{{a_,b_},{c_,d_}}]]:=
        {Line[{{a,b},(2/3)*{a,b} + (1/3)*{c,d}}],
         Line[{(2/3)*{a,b} + (1/3)*{c,d},
         (t/2)*{a - b + c + d,a + b - c + d} +
         ((1 - t)/2)*{a + b + c - d,-a + b + c + d}}],
         Line[{(t/2)*{a - b + c + d,a + b - c + d} +
         ((1 - t)/2)*{a + b + c - d,-a + b + c + d},
         (1/3)*{a,b} + (2/3)*{c,d}}],
        Line[{(1/3)*{a,b} + (2/3)*{c,d},{c,d}}]}
*)

tooth[t_][Line[{{a_,b_},{c_,d_}}]]:=
        {Line[{{a,b},(2/3)*{a,b} + (1/3)*{c,d}}],
         Line[{(2/3)*{a,b} + (1/3)*{c,d},
         {(a + c)/2 + t*(b - d),(b + d)/2 +t*(c - a)}}],
         Line[{{(a + c)/2 + t*(b - d),(b + d)/2 +t*(c - a)},
         (1/3)*{a,b} + (2/3)*{c,d}}],
        Line[{(1/3)*{a,b} + (2/3)*{c,d},{c,d}}]}

   kochsnowflake[m_]:=
    teeth[m,(3 + Sqrt[3])/6][linetolines[regularpolygon[3,1]]]

   kochsnowflake[m_,t_]:=
    teeth[m,t][linetolines[regularpolygon[3,1]]]
    
End[]

Protect["`*"];

EndPackage[]

