(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 9.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[    371453,       7898]
NotebookOptionsPosition[    365648,       7796]
NotebookOutlinePosition[    366012,       7812]
CellTagsIndexPosition[    365969,       7809]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell["Gauss Curvature", "Title",
 CellChangeTimes->{{3.7075707128745117`*^9, 
  3.707570723849238*^9}},ExpressionUUID->"facb1aa4-862e-467b-94f5-\
098711a93ba6"],

Cell[TextData[StyleBox["by  Rolf Sulanke",
 FontSize->24]], "Text",
 CellChangeTimes->{{3.707570738167346*^9, 
  3.707570771891259*^9}},ExpressionUUID->"339ff77c-1b17-45e7-95a4-\
cdad4aef9a2f"],

Cell[TextData[{
 "Created with ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " v. 9.0.1.0, June 27, 2017"
}], "Text",
 CellChangeTimes->{{3.707570803562709*^9, 
  3.707570851009181*^9}},ExpressionUUID->"86ae8b41-ccbd-4b75-b94d-\
adb18473354e"],

Cell["\<\
This is a Subnotebook of the notebook euklsfv4.nb, \
\[CloseCurlyDoubleQuote]Surfaces in the Euclidean 3-Space\
\[CloseCurlyDoubleQuote].
First open this notebook and run its initialization.
See in particular its subsection 2.2.1. There I calculated the Gauss \
curvature of a general unspecified surface sf[u,v]. The simplification of the \
obtained formula needed more than 3 hours. The result of the simplification \
is the kernel of the following program calculating the curvature directly \
from the three components of the surface.  Sometimes this Module calculates \
the curvature faster than other programs.\
\>", "Text",
 CellChangeTimes->{{3.707570860607684*^9, 3.707570932273308*^9}, {
   3.707570973217813*^9, 3.707571098575678*^9}, {3.7075712542323637`*^9, 
   3.707571588886485*^9}, {3.708321542199984*^9, 3.708321550687622*^9}, {
   3.708321597871688*^9, 3.70832161971146*^9}, 3.790429798189087*^9, {
   3.790429925373362*^9, 
   3.7904299281622953`*^9}},ExpressionUUID->"dea1b839-90da-4ec8-af4e-\
67fff5ea1a65"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"gausscurvature", "[", "ssf_", "]"}], "[", 
   RowBox[{"u1_", ",", "u2_"}], "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"f", ",", "i", ",", "u", ",", "v"}], "}"}], ",", 
    "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"f", "[", "i_", "]"}], "[", 
       RowBox[{"u_", ",", "v_"}], "]"}], ":=", 
      RowBox[{
       RowBox[{"ssf", "[", 
        RowBox[{"u", ",", "v"}], "]"}], "[", 
       RowBox[{"[", "i", "]"}], "]"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", 
           SuperscriptBox[
            RowBox[{
             SuperscriptBox[
              RowBox[{"f", "[", "3", "]"}], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"0", ",", "1"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{"u", ",", "v"}], "]"}], "2"]}], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{
             SuperscriptBox[
              RowBox[{
               SuperscriptBox[
                RowBox[{"f", "[", "2", "]"}], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"1", ",", "0"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{"u", ",", "v"}], "]"}], "2"], " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "1", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], "2"], "-", 
               RowBox[{
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "1", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], " ", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "1", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}]}]}], ")"}]}], "+", 
            RowBox[{
             RowBox[{
              SuperscriptBox[
               RowBox[{"f", "[", "1", "]"}], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"1", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{"u", ",", "v"}], "]"}], " ", 
             RowBox[{
              SuperscriptBox[
               RowBox[{"f", "[", "2", "]"}], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"1", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{"u", ",", "v"}], "]"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"-", "2"}], " ", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "1", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], " ", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "2", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}]}], "+", 
               RowBox[{
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "2", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], " ", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "1", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}]}], "+", 
               RowBox[{
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "1", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], " ", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "2", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}]}]}], ")"}]}], "+", 
            RowBox[{
             SuperscriptBox[
              RowBox[{
               SuperscriptBox[
                RowBox[{"f", "[", "1", "]"}], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"1", ",", "0"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{"u", ",", "v"}], "]"}], "2"], " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "2", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], "2"], "-", 
               RowBox[{
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "2", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], " ", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "2", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}]}]}], ")"}]}]}], ")"}]}], "-", 
         
         RowBox[{
          SuperscriptBox[
           RowBox[{
            SuperscriptBox[
             RowBox[{"f", "[", "2", "]"}], 
             TagBox[
              RowBox[{"(", 
               RowBox[{"0", ",", "1"}], ")"}],
              Derivative],
             MultilineFunction->None], "[", 
            RowBox[{"u", ",", "v"}], "]"}], "2"], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{
             SuperscriptBox[
              RowBox[{
               SuperscriptBox[
                RowBox[{"f", "[", "3", "]"}], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"1", ",", "0"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{"u", ",", "v"}], "]"}], "2"], " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "1", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], "2"], "-", 
               RowBox[{
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "1", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], " ", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "1", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}]}]}], ")"}]}], "+", 
            RowBox[{
             RowBox[{
              SuperscriptBox[
               RowBox[{"f", "[", "1", "]"}], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"1", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{"u", ",", "v"}], "]"}], " ", 
             RowBox[{
              SuperscriptBox[
               RowBox[{"f", "[", "3", "]"}], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"1", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{"u", ",", "v"}], "]"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"-", "2"}], " ", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "1", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], " ", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "3", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}]}], "+", 
               RowBox[{
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "3", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], " ", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "1", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}]}], "+", 
               RowBox[{
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "1", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], " ", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "3", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}]}]}], ")"}]}], "+", 
            RowBox[{
             SuperscriptBox[
              RowBox[{
               SuperscriptBox[
                RowBox[{"f", "[", "1", "]"}], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"1", ",", "0"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{"u", ",", "v"}], "]"}], "2"], " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "3", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], "2"], "-", 
               RowBox[{
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "3", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], " ", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "3", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}]}]}], ")"}]}]}], ")"}]}], "-", 
         
         RowBox[{
          SuperscriptBox[
           RowBox[{
            SuperscriptBox[
             RowBox[{"f", "[", "1", "]"}], 
             TagBox[
              RowBox[{"(", 
               RowBox[{"0", ",", "1"}], ")"}],
              Derivative],
             MultilineFunction->None], "[", 
            RowBox[{"u", ",", "v"}], "]"}], "2"], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{
             SuperscriptBox[
              RowBox[{
               SuperscriptBox[
                RowBox[{"f", "[", "3", "]"}], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"1", ",", "0"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{"u", ",", "v"}], "]"}], "2"], " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "2", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], "2"], "-", 
               RowBox[{
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "2", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], " ", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "2", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}]}]}], ")"}]}], "+", 
            RowBox[{
             RowBox[{
              SuperscriptBox[
               RowBox[{"f", "[", "2", "]"}], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"1", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{"u", ",", "v"}], "]"}], " ", 
             RowBox[{
              SuperscriptBox[
               RowBox[{"f", "[", "3", "]"}], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"1", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{"u", ",", "v"}], "]"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{"-", "2"}], " ", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "2", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], " ", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "3", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}]}], "+", 
               RowBox[{
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "3", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], " ", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "2", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}]}], "+", 
               RowBox[{
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "2", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], " ", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "3", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}]}]}], ")"}]}], "+", 
            RowBox[{
             SuperscriptBox[
              RowBox[{
               SuperscriptBox[
                RowBox[{"f", "[", "2", "]"}], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"1", ",", "0"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{"u", ",", "v"}], "]"}], "2"], " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "3", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], "2"], "-", 
               RowBox[{
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "3", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], " ", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "3", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}]}]}], ")"}]}]}], ")"}]}], "+", 
         
         RowBox[{
          RowBox[{
           SuperscriptBox[
            RowBox[{"f", "[", "1", "]"}], 
            TagBox[
             RowBox[{"(", 
              RowBox[{"0", ",", "1"}], ")"}],
             Derivative],
            MultilineFunction->None], "[", 
           RowBox[{"u", ",", "v"}], "]"}], " ", 
          RowBox[{
           SuperscriptBox[
            RowBox[{"f", "[", "2", "]"}], 
            TagBox[
             RowBox[{"(", 
              RowBox[{"0", ",", "1"}], ")"}],
             Derivative],
            MultilineFunction->None], "[", 
           RowBox[{"u", ",", "v"}], "]"}], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{
             SuperscriptBox[
              RowBox[{
               SuperscriptBox[
                RowBox[{"f", "[", "3", "]"}], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"1", ",", "0"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{"u", ",", "v"}], "]"}], "2"], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"2", " ", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "1", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], " ", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "2", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}]}], "-", 
               RowBox[{
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "2", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], " ", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "1", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}]}], "-", 
               RowBox[{
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "1", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], " ", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "2", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}]}]}], ")"}]}], "+", 
            RowBox[{"2", " ", 
             RowBox[{
              SuperscriptBox[
               RowBox[{"f", "[", "1", "]"}], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"1", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{"u", ",", "v"}], "]"}], " ", 
             RowBox[{
              SuperscriptBox[
               RowBox[{"f", "[", "2", "]"}], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"1", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{"u", ",", "v"}], "]"}], " ", 
             RowBox[{"(", 
              RowBox[{
               SuperscriptBox[
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "3", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], "2"], "-", 
               RowBox[{
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "3", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], " ", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "3", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}]}]}], ")"}]}], "+", 
            RowBox[{
             RowBox[{
              SuperscriptBox[
               RowBox[{"f", "[", "3", "]"}], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"1", ",", "0"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{"u", ",", "v"}], "]"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "2", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   RowBox[{"-", "2"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "1", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "3", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}]}], "+", 
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "3", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "1", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}]}], "+", 
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "1", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "3", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}]}]}], ")"}]}], "+", 
               RowBox[{
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "1", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   RowBox[{"-", "2"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "2", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "3", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}]}], "+", 
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "3", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "2", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}]}], "+", 
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "2", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "3", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}]}]}], ")"}]}]}], ")"}]}]}], 
           ")"}]}], "+", 
         RowBox[{
          RowBox[{
           SuperscriptBox[
            RowBox[{"f", "[", "3", "]"}], 
            TagBox[
             RowBox[{"(", 
              RowBox[{"0", ",", "1"}], ")"}],
             Derivative],
            MultilineFunction->None], "[", 
           RowBox[{"u", ",", "v"}], "]"}], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{
             RowBox[{
              SuperscriptBox[
               RowBox[{"f", "[", "2", "]"}], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"0", ",", "1"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{"u", ",", "v"}], "]"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "2", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"2", " ", 
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "3", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "1", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], "2"], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "1", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "1", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}]}]}], ")"}]}], "+", 
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "1", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "1", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "3", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "3", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "1", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "1", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "3", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}]}]}], ")"}]}]}], ")"}]}], 
               "+", 
               RowBox[{
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "1", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "3", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "1", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "2", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "2", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "1", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "1", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "2", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}]}]}], ")"}]}], "+", 
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "1", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{"2", " ", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "2", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "3", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "3", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "2", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}]}], "-", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "2", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "3", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}]}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}], "+", 
            RowBox[{
             RowBox[{
              SuperscriptBox[
               RowBox[{"f", "[", "1", "]"}], 
               TagBox[
                RowBox[{"(", 
                 RowBox[{"0", ",", "1"}], ")"}],
                Derivative],
               MultilineFunction->None], "[", 
              RowBox[{"u", ",", "v"}], "]"}], " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"2", " ", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "1", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], " ", 
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "3", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], " ", 
                RowBox[{"(", 
                 RowBox[{
                  SuperscriptBox[
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "2", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], "2"], "-", 
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "2", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "2", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}]}]}], ")"}]}], "+", 
               RowBox[{
                SuperscriptBox[
                 RowBox[{
                  SuperscriptBox[
                   RowBox[{"f", "[", "2", "]"}], 
                   TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                   MultilineFunction->None], "[", 
                  RowBox[{"u", ",", "v"}], "]"}], "2"], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"2", " ", 
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "1", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "3", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}]}], "-", 
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "3", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "1", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}]}], "-", 
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "1", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "3", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}]}]}], ")"}]}], "+", 
               RowBox[{
                RowBox[{
                 SuperscriptBox[
                  RowBox[{"f", "[", "2", "]"}], 
                  TagBox[
                   RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                   Derivative],
                  MultilineFunction->None], "[", 
                 RowBox[{"u", ",", "v"}], "]"}], " ", 
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "3", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "1", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "2", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "2", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "1", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "1", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "2", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}]}]}], ")"}]}], "+", 
                  RowBox[{
                   RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "1", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                   RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"-", "2"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "2", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "3", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"1", ",", "1"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "3", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "2", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}]}], "+", 
                    RowBox[{
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "2", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"0", ",", "2"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}], " ", 
                    RowBox[{
                    SuperscriptBox[
                    RowBox[{"f", "[", "3", "]"}], 
                    TagBox[
                    RowBox[{"(", 
                    RowBox[{"2", ",", "0"}], ")"}],
                    Derivative],
                    MultilineFunction->None], "[", 
                    RowBox[{"u", ",", "v"}], "]"}]}]}], ")"}]}]}], ")"}]}]}], 
              ")"}]}]}], ")"}]}]}], ")"}], "/", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           SuperscriptBox[
            RowBox[{
             SuperscriptBox[
              RowBox[{"f", "[", "3", "]"}], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"0", ",", "1"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{"u", ",", "v"}], "]"}], "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              RowBox[{
               SuperscriptBox[
                RowBox[{"f", "[", "1", "]"}], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"1", ",", "0"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{"u", ",", "v"}], "]"}], "2"], "+", 
             SuperscriptBox[
              RowBox[{
               SuperscriptBox[
                RowBox[{"f", "[", "2", "]"}], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"1", ",", "0"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{"u", ",", "v"}], "]"}], "2"]}], ")"}]}], "-", 
          RowBox[{"2", " ", 
           RowBox[{
            SuperscriptBox[
             RowBox[{"f", "[", "1", "]"}], 
             TagBox[
              RowBox[{"(", 
               RowBox[{"0", ",", "1"}], ")"}],
              Derivative],
             MultilineFunction->None], "[", 
            RowBox[{"u", ",", "v"}], "]"}], " ", 
           RowBox[{
            SuperscriptBox[
             RowBox[{"f", "[", "3", "]"}], 
             TagBox[
              RowBox[{"(", 
               RowBox[{"0", ",", "1"}], ")"}],
              Derivative],
             MultilineFunction->None], "[", 
            RowBox[{"u", ",", "v"}], "]"}], " ", 
           RowBox[{
            SuperscriptBox[
             RowBox[{"f", "[", "1", "]"}], 
             TagBox[
              RowBox[{"(", 
               RowBox[{"1", ",", "0"}], ")"}],
              Derivative],
             MultilineFunction->None], "[", 
            RowBox[{"u", ",", "v"}], "]"}], " ", 
           RowBox[{
            SuperscriptBox[
             RowBox[{"f", "[", "3", "]"}], 
             TagBox[
              RowBox[{"(", 
               RowBox[{"1", ",", "0"}], ")"}],
              Derivative],
             MultilineFunction->None], "[", 
            RowBox[{"u", ",", "v"}], "]"}]}], "-", 
          RowBox[{"2", " ", 
           RowBox[{
            SuperscriptBox[
             RowBox[{"f", "[", "2", "]"}], 
             TagBox[
              RowBox[{"(", 
               RowBox[{"0", ",", "1"}], ")"}],
              Derivative],
             MultilineFunction->None], "[", 
            RowBox[{"u", ",", "v"}], "]"}], " ", 
           RowBox[{
            SuperscriptBox[
             RowBox[{"f", "[", "2", "]"}], 
             TagBox[
              RowBox[{"(", 
               RowBox[{"1", ",", "0"}], ")"}],
              Derivative],
             MultilineFunction->None], "[", 
            RowBox[{"u", ",", "v"}], "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{
               SuperscriptBox[
                RowBox[{"f", "[", "1", "]"}], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"0", ",", "1"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{"u", ",", "v"}], "]"}], " ", 
              RowBox[{
               SuperscriptBox[
                RowBox[{"f", "[", "1", "]"}], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"1", ",", "0"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{"u", ",", "v"}], "]"}]}], "+", 
             RowBox[{
              RowBox[{
               SuperscriptBox[
                RowBox[{"f", "[", "3", "]"}], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"0", ",", "1"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{"u", ",", "v"}], "]"}], " ", 
              RowBox[{
               SuperscriptBox[
                RowBox[{"f", "[", "3", "]"}], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"1", ",", "0"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{"u", ",", "v"}], "]"}]}]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox[
            RowBox[{
             SuperscriptBox[
              RowBox[{"f", "[", "2", "]"}], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"0", ",", "1"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{"u", ",", "v"}], "]"}], "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              RowBox[{
               SuperscriptBox[
                RowBox[{"f", "[", "1", "]"}], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"1", ",", "0"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{"u", ",", "v"}], "]"}], "2"], "+", 
             SuperscriptBox[
              RowBox[{
               SuperscriptBox[
                RowBox[{"f", "[", "3", "]"}], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"1", ",", "0"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{"u", ",", "v"}], "]"}], "2"]}], ")"}]}], "+", 
          RowBox[{
           SuperscriptBox[
            RowBox[{
             SuperscriptBox[
              RowBox[{"f", "[", "1", "]"}], 
              TagBox[
               RowBox[{"(", 
                RowBox[{"0", ",", "1"}], ")"}],
               Derivative],
              MultilineFunction->None], "[", 
             RowBox[{"u", ",", "v"}], "]"}], "2"], " ", 
           RowBox[{"(", 
            RowBox[{
             SuperscriptBox[
              RowBox[{
               SuperscriptBox[
                RowBox[{"f", "[", "2", "]"}], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"1", ",", "0"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{"u", ",", "v"}], "]"}], "2"], "+", 
             SuperscriptBox[
              RowBox[{
               SuperscriptBox[
                RowBox[{"f", "[", "3", "]"}], 
                TagBox[
                 RowBox[{"(", 
                  RowBox[{"1", ",", "0"}], ")"}],
                 Derivative],
                MultilineFunction->None], "[", 
               RowBox[{"u", ",", "v"}], "]"}], "2"]}], ")"}]}]}], ")"}], 
        "2"]}], "/.", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"u", "\[Rule]", "u1"}], ",", 
        RowBox[{"v", "\[Rule]", "u2"}]}], "}"}]}]}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.707560131818461*^9, 3.7075601332252007`*^9}, 
   3.7075602378476677`*^9, {3.707560306716549*^9, 3.707560400644216*^9}, {
   3.707560485875721*^9, 3.707560798667779*^9}, {3.7075608344042*^9, 
   3.707560864303288*^9}, 
   3.70756113733438*^9},ExpressionUUID->"a2de38dc-5fbd-41dd-bd0b-\
9032f262b5fd"],

Cell[CellGroupData[{

Cell["Examples", "Section",
 CellChangeTimes->{{3.707571609150193*^9, 
  3.707571616325769*^9}},ExpressionUUID->"247e319f-dbbf-47bc-a839-\
ba778860ff55"],

Cell["The sphere", "Text",
 CellChangeTimes->{{3.70757168683628*^9, 
  3.707571692405181*^9}},ExpressionUUID->"ee6fbf9e-5083-40bb-a8f6-\
344d039722d4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"sphero", "[", "r", "]"}], "[", 
  RowBox[{"u", ",", "v"}], "]"}]], "Input",
 CellChangeTimes->{{3.707560422383925*^9, 
  3.707560432520459*^9}},ExpressionUUID->"4f47a352-3ecd-4f46-9f6b-\
b8f168c29eea"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"r", " ", 
    RowBox[{"Cos", "[", "u", "]"}], " ", 
    RowBox[{"Cos", "[", "v", "]"}]}], ",", 
   RowBox[{"r", " ", 
    RowBox[{"Cos", "[", "v", "]"}], " ", 
    RowBox[{"Sin", "[", "u", "]"}]}], ",", 
   RowBox[{"r", " ", 
    RowBox[{"Sin", "[", "v", "]"}]}]}], "}"}]], "Output",
 CellChangeTimes->{
  3.707560434198793*^9},ExpressionUUID->"5076a385-816d-45bc-beae-\
c875b6d47c2b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"gausscurvature", "[", 
   RowBox[{"sphero", "[", "r", "]"}], "]"}], "[", 
  RowBox[{"u", ",", "v"}], "]"}]], "Input",
 CellChangeTimes->{{3.707560871040197*^9, 3.707560912653556*^9}, {
  3.707561037507514*^9, 
  3.707561045508292*^9}},ExpressionUUID->"8a8834a5-a9b0-40df-abbe-\
19e9e154f133"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{
    RowBox[{"r", " ", 
     RowBox[{"Cos", "[", "v", "]"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "r"}], " ", 
        RowBox[{"Cos", "[", "u", "]"}], " ", 
        RowBox[{"Sin", "[", "v", "]"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", 
            SuperscriptBox["r", "4"]}], " ", 
           SuperscriptBox[
            RowBox[{"Cos", "[", "u", "]"}], "3"], " ", 
           SuperscriptBox[
            RowBox[{"Cos", "[", "v", "]"}], "3"], " ", 
           RowBox[{"Sin", "[", "v", "]"}]}], "-", 
          RowBox[{
           SuperscriptBox["r", "4"], " ", 
           RowBox[{"Cos", "[", "u", "]"}], " ", 
           SuperscriptBox[
            RowBox[{"Cos", "[", "v", "]"}], "3"], " ", 
           SuperscriptBox[
            RowBox[{"Sin", "[", "u", "]"}], "2"], " ", 
           RowBox[{"Sin", "[", "v", "]"}]}]}], ")"}]}], "-", 
       RowBox[{"r", " ", 
        RowBox[{"Sin", "[", "u", "]"}], " ", 
        RowBox[{"Sin", "[", "v", "]"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", 
            SuperscriptBox["r", "4"]}], " ", 
           SuperscriptBox[
            RowBox[{"Cos", "[", "u", "]"}], "2"], " ", 
           SuperscriptBox[
            RowBox[{"Cos", "[", "v", "]"}], "3"], " ", 
           RowBox[{"Sin", "[", "u", "]"}], " ", 
           RowBox[{"Sin", "[", "v", "]"}]}], "-", 
          RowBox[{
           SuperscriptBox["r", "4"], " ", 
           SuperscriptBox[
            RowBox[{"Cos", "[", "v", "]"}], "3"], " ", 
           SuperscriptBox[
            RowBox[{"Sin", "[", "u", "]"}], "3"], " ", 
           RowBox[{"Sin", "[", "v", "]"}]}]}], ")"}]}]}], ")"}]}], "-", 
    RowBox[{
     SuperscriptBox["r", "2"], " ", 
     SuperscriptBox[
      RowBox[{"Cos", "[", "v", "]"}], "2"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        SuperscriptBox["r", "2"], " ", 
        SuperscriptBox[
         RowBox[{"Cos", "[", "v", "]"}], "2"], " ", 
        SuperscriptBox[
         RowBox[{"Sin", "[", "u", "]"}], "2"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", 
            SuperscriptBox["r", "2"]}], " ", 
           SuperscriptBox[
            RowBox[{"Cos", "[", "v", "]"}], "2"], " ", 
           SuperscriptBox[
            RowBox[{"Sin", "[", "u", "]"}], "2"]}], "+", 
          RowBox[{
           SuperscriptBox["r", "2"], " ", 
           SuperscriptBox[
            RowBox[{"Cos", "[", "u", "]"}], "2"], " ", 
           SuperscriptBox[
            RowBox[{"Sin", "[", "v", "]"}], "2"]}]}], ")"}]}], "-", 
       RowBox[{
        SuperscriptBox["r", "2"], " ", 
        RowBox[{"Cos", "[", "u", "]"}], " ", 
        SuperscriptBox[
         RowBox[{"Cos", "[", "v", "]"}], "2"], " ", 
        RowBox[{"Sin", "[", "u", "]"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"2", " ", 
           SuperscriptBox["r", "2"], " ", 
           RowBox[{"Cos", "[", "u", "]"}], " ", 
           SuperscriptBox[
            RowBox[{"Cos", "[", "v", "]"}], "2"], " ", 
           RowBox[{"Sin", "[", "u", "]"}]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox["r", "2"], " ", 
           RowBox[{"Cos", "[", "u", "]"}], " ", 
           RowBox[{"Sin", "[", "u", "]"}], " ", 
           SuperscriptBox[
            RowBox[{"Sin", "[", "v", "]"}], "2"]}]}], ")"}]}], "+", 
       RowBox[{
        SuperscriptBox["r", "2"], " ", 
        SuperscriptBox[
         RowBox[{"Cos", "[", "u", "]"}], "2"], " ", 
        SuperscriptBox[
         RowBox[{"Cos", "[", "v", "]"}], "2"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", 
            SuperscriptBox["r", "2"]}], " ", 
           SuperscriptBox[
            RowBox[{"Cos", "[", "u", "]"}], "2"], " ", 
           SuperscriptBox[
            RowBox[{"Cos", "[", "v", "]"}], "2"]}], "+", 
          RowBox[{
           SuperscriptBox["r", "2"], " ", 
           SuperscriptBox[
            RowBox[{"Sin", "[", "u", "]"}], "2"], " ", 
           SuperscriptBox[
            RowBox[{"Sin", "[", "v", "]"}], "2"]}]}], ")"}]}]}], ")"}]}]}], 
   ")"}], "/", 
  SuperscriptBox[
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      SuperscriptBox["r", "2"], " ", 
      SuperscriptBox[
       RowBox[{"Cos", "[", "v", "]"}], "2"], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         SuperscriptBox["r", "2"], " ", 
         SuperscriptBox[
          RowBox[{"Cos", "[", "u", "]"}], "2"], " ", 
         SuperscriptBox[
          RowBox[{"Cos", "[", "v", "]"}], "2"]}], "+", 
        RowBox[{
         SuperscriptBox["r", "2"], " ", 
         SuperscriptBox[
          RowBox[{"Cos", "[", "v", "]"}], "2"], " ", 
         SuperscriptBox[
          RowBox[{"Sin", "[", "u", "]"}], "2"]}]}], ")"}]}], "+", 
     RowBox[{
      SuperscriptBox["r", "4"], " ", 
      SuperscriptBox[
       RowBox[{"Cos", "[", "u", "]"}], "4"], " ", 
      SuperscriptBox[
       RowBox[{"Cos", "[", "v", "]"}], "2"], " ", 
      SuperscriptBox[
       RowBox[{"Sin", "[", "v", "]"}], "2"]}], "+", 
     RowBox[{"2", " ", 
      SuperscriptBox["r", "4"], " ", 
      SuperscriptBox[
       RowBox[{"Cos", "[", "u", "]"}], "2"], " ", 
      SuperscriptBox[
       RowBox[{"Cos", "[", "v", "]"}], "2"], " ", 
      SuperscriptBox[
       RowBox[{"Sin", "[", "u", "]"}], "2"], " ", 
      SuperscriptBox[
       RowBox[{"Sin", "[", "v", "]"}], "2"]}], "+", 
     RowBox[{
      SuperscriptBox["r", "4"], " ", 
      SuperscriptBox[
       RowBox[{"Cos", "[", "v", "]"}], "2"], " ", 
      SuperscriptBox[
       RowBox[{"Sin", "[", "u", "]"}], "4"], " ", 
      SuperscriptBox[
       RowBox[{"Sin", "[", "v", "]"}], "2"]}]}], ")"}], "2"]}]], "Output",
 CellChangeTimes->{3.7075610466699877`*^9, 3.707561163384268*^9, 
  3.708321911194475*^9},ExpressionUUID->"7ec8ea6a-ae5b-47af-86b2-\
301df4b4cd14"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FullSimplify", "[", "%", "]"}]], "Input",
 CellChangeTimes->{{3.707560931518249*^9, 
  3.7075609428457327`*^9}},ExpressionUUID->"3b08beef-70a3-4a18-b042-\
90ca63468837"],

Cell[BoxData[
 FractionBox["1", 
  SuperscriptBox["r", "2"]]], "Output",
 CellChangeTimes->{3.7075609457150297`*^9, 3.707561076453884*^9, 
  3.7075611704750566`*^9, 
  3.708321934391303*^9},ExpressionUUID->"5839a286-4cb3-4ca1-8501-\
895ce68eadd4"]
}, Open  ]],

Cell["\[FilledSquare]", "Text",
 CellChangeTimes->{{3.707560931518249*^9, 3.7075609428457327`*^9}, 
   3.707561020939967*^9, {3.707571642303522*^9, 
   3.707571672149187*^9}},ExpressionUUID->"e55e5eac-1893-48ed-955b-\
aada8cdc0384"],

Cell["Tangent surfaces", "Text",
 CellChangeTimes->{{3.707571709145678*^9, 
  3.707571726038783*^9}},ExpressionUUID->"808252ce-4067-4535-8750-\
8412d9c4d59e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"curve", "[", "v", "]"}]], "Input",
 CellChangeTimes->{{3.707561225839074*^9, 
  3.7075612339159822`*^9}},ExpressionUUID->"1c36da4b-67e8-420c-87ea-\
9a64dd8c4df0"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"x", "[", "v", "]"}], "[", "1", "]"}], ",", 
   RowBox[{
    RowBox[{"x", "[", "v", "]"}], "[", "2", "]"}], ",", 
   RowBox[{
    RowBox[{"x", "[", "v", "]"}], "[", "3", "]"}]}], "}"}]], "Output",
 CellChangeTimes->{3.7075612350090523`*^9, 
  3.708321946464095*^9},ExpressionUUID->"166d12cb-5210-4947-b003-\
499eaa36914a"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"tang", "[", 
   RowBox[{"u_", ",", "v_"}], "]"}], ":=", " ", 
  RowBox[{
   RowBox[{"curve", "[", "v", "]"}], "+", 
   RowBox[{"u", "*", 
    RowBox[{
     RowBox[{"curve", "'"}], "[", "v", "]"}]}]}]}]], "Input",
 CellChangeTimes->{{3.707561251152027*^9, 3.707561328612564*^9}, 
   3.707561360917665*^9},ExpressionUUID->"1f182eeb-66f9-4af4-bc26-\
61eaa99cc547"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"tang", "[", 
  RowBox[{"u", ",", "v"}], "]"}]], "Input",
 CellChangeTimes->{{3.7075613682033243`*^9, 
  3.707561375147464*^9}},ExpressionUUID->"9f9f2fed-a412-4011-95a8-\
5c2188ed9c38"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"x", "[", "v", "]"}], "[", "1", "]"}], "+", 
    RowBox[{"u", " ", 
     RowBox[{
      RowBox[{
       SuperscriptBox["x", "\[Prime]",
        MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"x", "[", "v", "]"}], "[", "2", "]"}], "+", 
    RowBox[{"u", " ", 
     RowBox[{
      RowBox[{
       SuperscriptBox["x", "\[Prime]",
        MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"x", "[", "v", "]"}], "[", "3", "]"}], "+", 
    RowBox[{"u", " ", 
     RowBox[{
      RowBox[{
       SuperscriptBox["x", "\[Prime]",
        MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}]}]}]}], 
  "}"}]], "Output",
 CellChangeTimes->{3.707561376302082*^9, 
  3.708321968895225*^9},ExpressionUUID->"c4955467-5c8d-49b2-93ba-\
30bb1e26c9cc"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"gausscurvature", "[", "tang", "]"}], "[", 
  RowBox[{"u", ",", "v"}], "]"}]], "Input",
 CellChangeTimes->{{3.707561385402895*^9, 
  3.707561411161817*^9}},ExpressionUUID->"35d99c3c-3337-45e1-bea6-\
6e110e948b16"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         SuperscriptBox[
          RowBox[{
           RowBox[{
            SuperscriptBox["x", "\[Prime]",
             MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}], "2"],
          " ", 
         SuperscriptBox[
          RowBox[{
           RowBox[{
            SuperscriptBox["x", "\[Prime]\[Prime]",
             MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], 
          "2"]}], "-", 
        RowBox[{"2", " ", 
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], " ", 
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}], " ", 
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], " ", 
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}]}], "+", 
        RowBox[{
         SuperscriptBox[
          RowBox[{
           RowBox[{
            SuperscriptBox["x", "\[Prime]",
             MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], "2"],
          " ", 
         SuperscriptBox[
          RowBox[{
           RowBox[{
            SuperscriptBox["x", "\[Prime]\[Prime]",
             MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}], 
          "2"]}]}], ")"}]}], " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{
          SuperscriptBox["x", "\[Prime]",
           MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}], "+", 
        RowBox[{"u", " ", 
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}]}]}], 
       ")"}], "2"]}], "-", 
    RowBox[{
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{
          SuperscriptBox["x", "\[Prime]",
           MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}], "+", 
        RowBox[{"u", " ", 
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}]}]}], 
       ")"}], "2"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        SuperscriptBox[
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}], "2"], 
        " ", 
        SuperscriptBox[
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], 
         "2"]}], "-", 
       RowBox[{"2", " ", 
        RowBox[{
         RowBox[{
          SuperscriptBox["x", "\[Prime]",
           MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], " ", 
        RowBox[{
         RowBox[{
          SuperscriptBox["x", "\[Prime]",
           MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}], " ", 
        RowBox[{
         RowBox[{
          SuperscriptBox["x", "\[Prime]\[Prime]",
           MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], " ", 
        RowBox[{
         RowBox[{
          SuperscriptBox["x", "\[Prime]\[Prime]",
           MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}]}], "+", 
       
       RowBox[{
        SuperscriptBox[
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], "2"], 
        " ", 
        SuperscriptBox[
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}], 
         "2"]}]}], ")"}]}], "-", 
    RowBox[{
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{
          SuperscriptBox["x", "\[Prime]",
           MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], "+", 
        RowBox[{"u", " ", 
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}]}]}], 
       ")"}], "2"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        SuperscriptBox[
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}], "2"], 
        " ", 
        SuperscriptBox[
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}], 
         "2"]}], "-", 
       RowBox[{"2", " ", 
        RowBox[{
         RowBox[{
          SuperscriptBox["x", "\[Prime]",
           MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}], " ", 
        RowBox[{
         RowBox[{
          SuperscriptBox["x", "\[Prime]",
           MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}], " ", 
        RowBox[{
         RowBox[{
          SuperscriptBox["x", "\[Prime]\[Prime]",
           MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}], " ", 
        RowBox[{
         RowBox[{
          SuperscriptBox["x", "\[Prime]\[Prime]",
           MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}]}], "+", 
       
       RowBox[{
        SuperscriptBox[
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}], "2"], 
        " ", 
        SuperscriptBox[
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}], 
         "2"]}]}], ")"}]}], "+", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{
         SuperscriptBox["x", "\[Prime]",
          MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], "+", 
       RowBox[{"u", " ", 
        RowBox[{
         RowBox[{
          SuperscriptBox["x", "\[Prime]\[Prime]",
           MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}]}]}], 
      ")"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{
         SuperscriptBox["x", "\[Prime]",
          MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}], "+", 
       RowBox[{"u", " ", 
        RowBox[{
         RowBox[{
          SuperscriptBox["x", "\[Prime]\[Prime]",
           MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}]}]}], 
      ")"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"2", " ", 
        SuperscriptBox[
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}], "2"], 
        " ", 
        RowBox[{
         RowBox[{
          SuperscriptBox["x", "\[Prime]\[Prime]",
           MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], " ", 
        RowBox[{
         RowBox[{
          SuperscriptBox["x", "\[Prime]\[Prime]",
           MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}]}], "+", 
       
       RowBox[{"2", " ", 
        RowBox[{
         RowBox[{
          SuperscriptBox["x", "\[Prime]",
           MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], " ", 
        RowBox[{
         RowBox[{
          SuperscriptBox["x", "\[Prime]",
           MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}], " ", 
        SuperscriptBox[
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}], 
         "2"]}], "+", 
       RowBox[{
        RowBox[{
         RowBox[{
          SuperscriptBox["x", "\[Prime]",
           MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "2"}], " ", 
           RowBox[{
            RowBox[{
             SuperscriptBox["x", "\[Prime]",
              MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}], " ", 
           RowBox[{
            RowBox[{
             SuperscriptBox["x", "\[Prime]\[Prime]",
              MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], " ", 
           RowBox[{
            RowBox[{
             SuperscriptBox["x", "\[Prime]\[Prime]",
              MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}]}], 
          "-", 
          RowBox[{"2", " ", 
           RowBox[{
            RowBox[{
             SuperscriptBox["x", "\[Prime]",
              MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], " ", 
           RowBox[{
            RowBox[{
             SuperscriptBox["x", "\[Prime]\[Prime]",
              MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}], " ", 
           RowBox[{
            RowBox[{
             SuperscriptBox["x", "\[Prime]\[Prime]",
              MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}]}]}], 
         ")"}]}]}], ")"}]}], "+", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{
         SuperscriptBox["x", "\[Prime]",
          MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}], "+", 
       RowBox[{"u", " ", 
        RowBox[{
         RowBox[{
          SuperscriptBox["x", "\[Prime]\[Prime]",
           MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}]}]}], 
      ")"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{
            SuperscriptBox["x", "\[Prime]",
             MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], "+", 
          
          RowBox[{"u", " ", 
           RowBox[{
            RowBox[{
             SuperscriptBox["x", "\[Prime]\[Prime]",
              MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}]}]}], 
         ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"2", " ", 
           RowBox[{
            RowBox[{
             SuperscriptBox["x", "\[Prime]",
              MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], " ", 
           RowBox[{
            RowBox[{
             SuperscriptBox["x", "\[Prime]",
              MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}], " ", 
           SuperscriptBox[
            RowBox[{
             RowBox[{
              SuperscriptBox["x", "\[Prime]\[Prime]",
               MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}], 
            "2"]}], "+", 
          RowBox[{"2", " ", 
           SuperscriptBox[
            RowBox[{
             RowBox[{
              SuperscriptBox["x", "\[Prime]",
               MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}], 
            "2"], " ", 
           RowBox[{
            RowBox[{
             SuperscriptBox["x", "\[Prime]\[Prime]",
              MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], " ", 
           RowBox[{
            RowBox[{
             SuperscriptBox["x", "\[Prime]\[Prime]",
              MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}]}], 
          "+", 
          RowBox[{
           RowBox[{
            RowBox[{
             SuperscriptBox["x", "\[Prime]",
              MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "2"}], " ", 
              RowBox[{
               RowBox[{
                SuperscriptBox["x", "\[Prime]",
                 MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}], 
              " ", 
              RowBox[{
               RowBox[{
                SuperscriptBox["x", "\[Prime]\[Prime]",
                 MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], 
              " ", 
              RowBox[{
               RowBox[{
                SuperscriptBox["x", "\[Prime]\[Prime]",
                 MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}]}],
              "-", 
             RowBox[{"2", " ", 
              RowBox[{
               RowBox[{
                SuperscriptBox["x", "\[Prime]",
                 MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], 
              " ", 
              RowBox[{
               RowBox[{
                SuperscriptBox["x", "\[Prime]\[Prime]",
                 MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}], 
              " ", 
              RowBox[{
               RowBox[{
                SuperscriptBox["x", "\[Prime]\[Prime]",
                 MultilineFunction->None], "[", "v", "]"}], "[", "3", 
               "]"}]}]}], ")"}]}]}], ")"}]}], "+", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{
            SuperscriptBox["x", "\[Prime]",
             MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}], "+", 
          
          RowBox[{"u", " ", 
           RowBox[{
            RowBox[{
             SuperscriptBox["x", "\[Prime]\[Prime]",
              MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}]}]}], 
         ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{
            RowBox[{
             SuperscriptBox["x", "\[Prime]",
              MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", 
              RowBox[{
               RowBox[{
                SuperscriptBox["x", "\[Prime]",
                 MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}], 
              " ", 
              SuperscriptBox[
               RowBox[{
                RowBox[{
                 SuperscriptBox["x", "\[Prime]\[Prime]",
                  MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], 
               "2"]}], "-", 
             RowBox[{"2", " ", 
              RowBox[{
               RowBox[{
                SuperscriptBox["x", "\[Prime]",
                 MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], 
              " ", 
              RowBox[{
               RowBox[{
                SuperscriptBox["x", "\[Prime]\[Prime]",
                 MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], 
              " ", 
              RowBox[{
               RowBox[{
                SuperscriptBox["x", "\[Prime]\[Prime]",
                 MultilineFunction->None], "[", "v", "]"}], "[", "3", 
               "]"}]}]}], ")"}]}], "+", 
          RowBox[{
           RowBox[{
            RowBox[{
             SuperscriptBox["x", "\[Prime]",
              MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "2"}], " ", 
              RowBox[{
               RowBox[{
                SuperscriptBox["x", "\[Prime]",
                 MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}], 
              " ", 
              RowBox[{
               RowBox[{
                SuperscriptBox["x", "\[Prime]\[Prime]",
                 MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], 
              " ", 
              RowBox[{
               RowBox[{
                SuperscriptBox["x", "\[Prime]\[Prime]",
                 MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}]}],
              "+", 
             RowBox[{"2", " ", 
              RowBox[{
               RowBox[{
                SuperscriptBox["x", "\[Prime]",
                 MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], 
              " ", 
              RowBox[{
               RowBox[{
                SuperscriptBox["x", "\[Prime]\[Prime]",
                 MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}], 
              " ", 
              RowBox[{
               RowBox[{
                SuperscriptBox["x", "\[Prime]\[Prime]",
                 MultilineFunction->None], "[", "v", "]"}], "[", "3", 
               "]"}]}]}], ")"}]}]}], ")"}]}]}], ")"}]}]}], ")"}], "/", 
  SuperscriptBox[
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox[
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}], "2"], 
        "+", 
        SuperscriptBox[
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}], 
         "2"]}], ")"}], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], "+", 
         RowBox[{"u", " ", 
          RowBox[{
           RowBox[{
            SuperscriptBox["x", "\[Prime]\[Prime]",
             MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}]}]}], 
        ")"}], "2"]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox[
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], "2"], 
        "+", 
        SuperscriptBox[
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}], 
         "2"]}], ")"}], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}], "+", 
         RowBox[{"u", " ", 
          RowBox[{
           RowBox[{
            SuperscriptBox["x", "\[Prime]\[Prime]",
             MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}]}]}], 
        ")"}], "2"]}], "-", 
     RowBox[{"2", " ", 
      RowBox[{
       RowBox[{
        SuperscriptBox["x", "\[Prime]",
         MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], " ", 
      RowBox[{
       RowBox[{
        SuperscriptBox["x", "\[Prime]",
         MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{
          SuperscriptBox["x", "\[Prime]",
           MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], "+", 
        RowBox[{"u", " ", 
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}]}]}], 
       ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{
          SuperscriptBox["x", "\[Prime]",
           MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}], "+", 
        RowBox[{"u", " ", 
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}]}]}], 
       ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox[
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], "2"], 
        "+", 
        SuperscriptBox[
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}], 
         "2"]}], ")"}], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}], "+", 
         RowBox[{"u", " ", 
          RowBox[{
           RowBox[{
            SuperscriptBox["x", "\[Prime]\[Prime]",
             MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}]}]}], 
        ")"}], "2"]}], "-", 
     RowBox[{"2", " ", 
      RowBox[{
       RowBox[{
        SuperscriptBox["x", "\[Prime]",
         MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{
          SuperscriptBox["x", "\[Prime]",
           MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}], "+", 
        RowBox[{"u", " ", 
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "2", "]"}]}]}], 
       ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{
             SuperscriptBox["x", "\[Prime]",
              MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}], "+", 
           RowBox[{"u", " ", 
            RowBox[{
             RowBox[{
              SuperscriptBox["x", "\[Prime]\[Prime]",
               MultilineFunction->None], "[", "v", "]"}], "[", "1", "]"}]}]}],
           ")"}]}], "+", 
        RowBox[{
         RowBox[{
          RowBox[{
           SuperscriptBox["x", "\[Prime]",
            MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{
             SuperscriptBox["x", "\[Prime]",
              MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}], "+", 
           RowBox[{"u", " ", 
            RowBox[{
             RowBox[{
              SuperscriptBox["x", "\[Prime]\[Prime]",
               MultilineFunction->None], "[", "v", "]"}], "[", "3", "]"}]}]}],
           ")"}]}]}], ")"}]}]}], ")"}], "2"]}]], "Output",
 CellChangeTimes->{3.707561412415298*^9, 
  3.7083219851676064`*^9},ExpressionUUID->"2093f016-b8f7-49a9-9ee6-\
ed89f04eef28"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Timing", "[", 
  RowBox[{"FullSimplify", "[", 
   RowBox[{"%", ",", 
    RowBox[{"TimeConstraint", " ", "\[Rule]", " ", "Infinity"}]}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.707561422512939*^9, 3.70756143829011*^9}, {
   3.707562115748949*^9, 3.707562214893868*^9}, 
   3.708322002619979*^9},ExpressionUUID->"be483a63-36f5-421a-85ed-\
53ff52ae5c26"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.004`", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.707561528142868*^9, 3.7075669009764853`*^9, 
  3.7083220070118923`*^9},ExpressionUUID->"3d72f217-b81a-4903-a977-\
5409191bc2f9"]
}, Open  ]],

Cell["Result 0 after  0.004 s with Mathematica v. 11.1.1.0.", "Text",
 CellChangeTimes->{{3.70832210301003*^9, 
  3.708322155266038*^9}},ExpressionUUID->"4dca132c-2b2c-4387-9bf1-\
5ca997af2d68"],

Cell["", "Text",
 CellChangeTimes->{{3.707566926064529*^9, 3.707566955015437*^9}, 
   3.7083216626059*^9, 
   3.708322073210567*^9},ExpressionUUID->"73404bc8-7cf7-4ac1-a95f-\
494e17a28896"],

Cell["\<\
\[FilledSquare]
Conical surfaces\
\>", "Text",
 CellChangeTimes->{{3.7075717633844748`*^9, 
  3.707571796803787*^9}},ExpressionUUID->"83a32c95-9c8f-4cf5-b7cf-\
01df489de978"],

Cell[BoxData[
 RowBox[{
  RowBox[{"conef", "[", 
   RowBox[{"u_", ",", "v_"}], "]"}], ":=", " ", 
  RowBox[{"u", "*", 
   RowBox[{
    RowBox[{"sphero", "[", "1", "]"}], "[", 
    RowBox[{
     RowBox[{"ff", "[", "v", "]"}], ",", 
     RowBox[{"gg", "[", "v", "]"}]}], "]"}]}]}]], "Input",
 CellChangeTimes->{{3.7075615329355516`*^9, 3.707561611681172*^9}, {
   3.7075627204716873`*^9, 3.707562770743494*^9}, 
   3.707566983292492*^9},ExpressionUUID->"6d5c337e-7648-4965-a07b-\
95def1ba1015"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"gausscurvature", "[", "conef", "]"}], "[", 
  RowBox[{"u", ",", "v"}], "]"}]], "Input",
 CellChangeTimes->{{3.70756278458606*^9, 3.707562814437171*^9}, 
   3.7075669952745743`*^9},ExpressionUUID->"ee3680d1-f5b2-4900-b550-\
af3344d7fac6"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         RowBox[{"u", " ", 
          RowBox[{"Cos", "[", 
           RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
          RowBox[{"Cos", "[", 
           RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
          RowBox[{
           SuperscriptBox["ff", "\[Prime]",
            MultilineFunction->None], "[", "v", "]"}]}], "-", 
         RowBox[{"u", " ", 
          RowBox[{"Sin", "[", 
           RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
          RowBox[{"Sin", "[", 
           RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
          RowBox[{
           SuperscriptBox["gg", "\[Prime]",
            MultilineFunction->None], "[", "v", "]"}]}]}], ")"}], "2"]}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        SuperscriptBox[
         RowBox[{"Cos", "[", 
          RowBox[{"ff", "[", "v", "]"}], "]"}], "2"], " ", 
        SuperscriptBox[
         RowBox[{"Cos", "[", 
          RowBox[{"gg", "[", "v", "]"}], "]"}], "4"], " ", 
        SuperscriptBox[
         RowBox[{
          SuperscriptBox["gg", "\[Prime]",
           MultilineFunction->None], "[", "v", "]"}], "2"]}], "-", 
       RowBox[{"2", " ", 
        RowBox[{"Cos", "[", 
         RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
        SuperscriptBox[
         RowBox[{"Cos", "[", 
          RowBox[{"gg", "[", "v", "]"}], "]"}], "2"], " ", 
        RowBox[{"Sin", "[", 
         RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
        RowBox[{
         SuperscriptBox["gg", "\[Prime]",
          MultilineFunction->None], "[", "v", "]"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", 
            RowBox[{"Cos", "[", 
             RowBox[{"gg", "[", "v", "]"}], "]"}]}], " ", 
           RowBox[{"Sin", "[", 
            RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
           RowBox[{
            SuperscriptBox["ff", "\[Prime]",
             MultilineFunction->None], "[", "v", "]"}]}], "-", 
          RowBox[{
           RowBox[{"Cos", "[", 
            RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
           RowBox[{"Sin", "[", 
            RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
           RowBox[{
            SuperscriptBox["gg", "\[Prime]",
             MultilineFunction->None], "[", "v", "]"}]}]}], ")"}]}], "+", 
       RowBox[{
        SuperscriptBox[
         RowBox[{"Sin", "[", 
          RowBox[{"gg", "[", "v", "]"}], "]"}], "2"], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", 
             RowBox[{"Cos", "[", 
              RowBox[{"gg", "[", "v", "]"}], "]"}]}], " ", 
            RowBox[{"Sin", "[", 
             RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
            RowBox[{
             SuperscriptBox["ff", "\[Prime]",
              MultilineFunction->None], "[", "v", "]"}]}], "-", 
           RowBox[{
            RowBox[{"Cos", "[", 
             RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
            RowBox[{"Sin", "[", 
             RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
            RowBox[{
             SuperscriptBox["gg", "\[Prime]",
              MultilineFunction->None], "[", "v", "]"}]}]}], ")"}], "2"]}]}], 
      ")"}]}], "-", 
    RowBox[{
     SuperscriptBox["u", "2"], " ", 
     SuperscriptBox[
      RowBox[{"Cos", "[", 
       RowBox[{"gg", "[", "v", "]"}], "]"}], "2"], " ", 
     SuperscriptBox[
      RowBox[{
       SuperscriptBox["gg", "\[Prime]",
        MultilineFunction->None], "[", "v", "]"}], "2"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        SuperscriptBox[
         RowBox[{"Cos", "[", 
          RowBox[{"gg", "[", "v", "]"}], "]"}], "2"], " ", 
        SuperscriptBox[
         RowBox[{"Sin", "[", 
          RowBox[{"ff", "[", "v", "]"}], "]"}], "2"], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", 
             RowBox[{"Cos", "[", 
              RowBox[{"gg", "[", "v", "]"}], "]"}]}], " ", 
            RowBox[{"Sin", "[", 
             RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
            RowBox[{
             SuperscriptBox["ff", "\[Prime]",
              MultilineFunction->None], "[", "v", "]"}]}], "-", 
           RowBox[{
            RowBox[{"Cos", "[", 
             RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
            RowBox[{"Sin", "[", 
             RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
            RowBox[{
             SuperscriptBox["gg", "\[Prime]",
              MultilineFunction->None], "[", "v", "]"}]}]}], ")"}], "2"]}], 
       "-", 
       RowBox[{"2", " ", 
        RowBox[{"Cos", "[", 
         RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
        SuperscriptBox[
         RowBox[{"Cos", "[", 
          RowBox[{"gg", "[", "v", "]"}], "]"}], "2"], " ", 
        RowBox[{"Sin", "[", 
         RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", 
            RowBox[{"Cos", "[", 
             RowBox[{"gg", "[", "v", "]"}], "]"}]}], " ", 
           RowBox[{"Sin", "[", 
            RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
           RowBox[{
            SuperscriptBox["ff", "\[Prime]",
             MultilineFunction->None], "[", "v", "]"}]}], "-", 
          RowBox[{
           RowBox[{"Cos", "[", 
            RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
           RowBox[{"Sin", "[", 
            RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
           RowBox[{
            SuperscriptBox["gg", "\[Prime]",
             MultilineFunction->None], "[", "v", "]"}]}]}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"Cos", "[", 
            RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
           RowBox[{"Cos", "[", 
            RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
           RowBox[{
            SuperscriptBox["ff", "\[Prime]",
             MultilineFunction->None], "[", "v", "]"}]}], "-", 
          RowBox[{
           RowBox[{"Sin", "[", 
            RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
           RowBox[{"Sin", "[", 
            RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
           RowBox[{
            SuperscriptBox["gg", "\[Prime]",
             MultilineFunction->None], "[", "v", "]"}]}]}], ")"}]}], "+", 
       RowBox[{
        SuperscriptBox[
         RowBox[{"Cos", "[", 
          RowBox[{"ff", "[", "v", "]"}], "]"}], "2"], " ", 
        SuperscriptBox[
         RowBox[{"Cos", "[", 
          RowBox[{"gg", "[", "v", "]"}], "]"}], "2"], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"Cos", "[", 
             RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
            RowBox[{"Cos", "[", 
             RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
            RowBox[{
             SuperscriptBox["ff", "\[Prime]",
              MultilineFunction->None], "[", "v", "]"}]}], "-", 
           RowBox[{
            RowBox[{"Sin", "[", 
             RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
            RowBox[{"Sin", "[", 
             RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
            RowBox[{
             SuperscriptBox["gg", "\[Prime]",
              MultilineFunction->None], "[", "v", "]"}]}]}], ")"}], "2"]}]}], 
      ")"}]}], "-", 
    RowBox[{
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "u"}], " ", 
         RowBox[{"Cos", "[", 
          RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
         RowBox[{"Sin", "[", 
          RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
         RowBox[{
          SuperscriptBox["ff", "\[Prime]",
           MultilineFunction->None], "[", "v", "]"}]}], "-", 
        RowBox[{"u", " ", 
         RowBox[{"Cos", "[", 
          RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
         RowBox[{"Sin", "[", 
          RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
         RowBox[{
          SuperscriptBox["gg", "\[Prime]",
           MultilineFunction->None], "[", "v", "]"}]}]}], ")"}], "2"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        SuperscriptBox[
         RowBox[{"Cos", "[", 
          RowBox[{"gg", "[", "v", "]"}], "]"}], "4"], " ", 
        SuperscriptBox[
         RowBox[{"Sin", "[", 
          RowBox[{"ff", "[", "v", "]"}], "]"}], "2"], " ", 
        SuperscriptBox[
         RowBox[{
          SuperscriptBox["gg", "\[Prime]",
           MultilineFunction->None], "[", "v", "]"}], "2"]}], "-", 
       RowBox[{"2", " ", 
        SuperscriptBox[
         RowBox[{"Cos", "[", 
          RowBox[{"gg", "[", "v", "]"}], "]"}], "2"], " ", 
        RowBox[{"Sin", "[", 
         RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
        RowBox[{"Sin", "[", 
         RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
        RowBox[{
         SuperscriptBox["gg", "\[Prime]",
          MultilineFunction->None], "[", "v", "]"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"Cos", "[", 
            RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
           RowBox[{"Cos", "[", 
            RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
           RowBox[{
            SuperscriptBox["ff", "\[Prime]",
             MultilineFunction->None], "[", "v", "]"}]}], "-", 
          RowBox[{
           RowBox[{"Sin", "[", 
            RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
           RowBox[{"Sin", "[", 
            RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
           RowBox[{
            SuperscriptBox["gg", "\[Prime]",
             MultilineFunction->None], "[", "v", "]"}]}]}], ")"}]}], "+", 
       RowBox[{
        SuperscriptBox[
         RowBox[{"Sin", "[", 
          RowBox[{"gg", "[", "v", "]"}], "]"}], "2"], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"Cos", "[", 
             RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
            RowBox[{"Cos", "[", 
             RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
            RowBox[{
             SuperscriptBox["ff", "\[Prime]",
              MultilineFunction->None], "[", "v", "]"}]}], "-", 
           RowBox[{
            RowBox[{"Sin", "[", 
             RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
            RowBox[{"Sin", "[", 
             RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
            RowBox[{
             SuperscriptBox["gg", "\[Prime]",
              MultilineFunction->None], "[", "v", "]"}]}]}], ")"}], "2"]}]}], 
      ")"}]}], "+", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "u"}], " ", 
        RowBox[{"Cos", "[", 
         RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
        RowBox[{"Sin", "[", 
         RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
        RowBox[{
         SuperscriptBox["ff", "\[Prime]",
          MultilineFunction->None], "[", "v", "]"}]}], "-", 
       RowBox[{"u", " ", 
        RowBox[{"Cos", "[", 
         RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
        RowBox[{"Sin", "[", 
         RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
        RowBox[{
         SuperscriptBox["gg", "\[Prime]",
          MultilineFunction->None], "[", "v", "]"}]}]}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"u", " ", 
        RowBox[{"Cos", "[", 
         RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
        RowBox[{"Cos", "[", 
         RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
        RowBox[{
         SuperscriptBox["ff", "\[Prime]",
          MultilineFunction->None], "[", "v", "]"}]}], "-", 
       RowBox[{"u", " ", 
        RowBox[{"Sin", "[", 
         RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
        RowBox[{"Sin", "[", 
         RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
        RowBox[{
         SuperscriptBox["gg", "\[Prime]",
          MultilineFunction->None], "[", "v", "]"}]}]}], ")"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"2", " ", 
        RowBox[{"Cos", "[", 
         RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
        SuperscriptBox[
         RowBox[{"Cos", "[", 
          RowBox[{"gg", "[", "v", "]"}], "]"}], "4"], " ", 
        RowBox[{"Sin", "[", 
         RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
        SuperscriptBox[
         RowBox[{
          SuperscriptBox["gg", "\[Prime]",
           MultilineFunction->None], "[", "v", "]"}], "2"]}], "+", 
       RowBox[{"2", " ", 
        SuperscriptBox[
         RowBox[{"Sin", "[", 
          RowBox[{"gg", "[", "v", "]"}], "]"}], "2"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", 
            RowBox[{"Cos", "[", 
             RowBox[{"gg", "[", "v", "]"}], "]"}]}], " ", 
           RowBox[{"Sin", "[", 
            RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
           RowBox[{
            SuperscriptBox["ff", "\[Prime]",
             MultilineFunction->None], "[", "v", "]"}]}], "-", 
          RowBox[{
           RowBox[{"Cos", "[", 
            RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
           RowBox[{"Sin", "[", 
            RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
           RowBox[{
            SuperscriptBox["gg", "\[Prime]",
             MultilineFunction->None], "[", "v", "]"}]}]}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"Cos", "[", 
            RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
           RowBox[{"Cos", "[", 
            RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
           RowBox[{
            SuperscriptBox["ff", "\[Prime]",
             MultilineFunction->None], "[", "v", "]"}]}], "-", 
          RowBox[{
           RowBox[{"Sin", "[", 
            RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
           RowBox[{"Sin", "[", 
            RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
           RowBox[{
            SuperscriptBox["gg", "\[Prime]",
             MultilineFunction->None], "[", "v", "]"}]}]}], ")"}]}], "+", 
       RowBox[{
        RowBox[{"Sin", "[", 
         RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "2"}], " ", 
           SuperscriptBox[
            RowBox[{"Cos", "[", 
             RowBox[{"gg", "[", "v", "]"}], "]"}], "2"], " ", 
           RowBox[{"Sin", "[", 
            RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
           RowBox[{
            SuperscriptBox["gg", "\[Prime]",
             MultilineFunction->None], "[", "v", "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", 
               RowBox[{"Cos", "[", 
                RowBox[{"gg", "[", "v", "]"}], "]"}]}], " ", 
              RowBox[{"Sin", "[", 
               RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
              RowBox[{
               SuperscriptBox["ff", "\[Prime]",
                MultilineFunction->None], "[", "v", "]"}]}], "-", 
             RowBox[{
              RowBox[{"Cos", "[", 
               RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
              RowBox[{"Sin", "[", 
               RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
              RowBox[{
               SuperscriptBox["gg", "\[Prime]",
                MultilineFunction->None], "[", "v", "]"}]}]}], ")"}]}], "-", 
          RowBox[{"2", " ", 
           RowBox[{"Cos", "[", 
            RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
           SuperscriptBox[
            RowBox[{"Cos", "[", 
             RowBox[{"gg", "[", "v", "]"}], "]"}], "2"], " ", 
           RowBox[{
            SuperscriptBox["gg", "\[Prime]",
             MultilineFunction->None], "[", "v", "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"Cos", "[", 
               RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
              RowBox[{"Cos", "[", 
               RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
              RowBox[{
               SuperscriptBox["ff", "\[Prime]",
                MultilineFunction->None], "[", "v", "]"}]}], "-", 
             RowBox[{
              RowBox[{"Sin", "[", 
               RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
              RowBox[{"Sin", "[", 
               RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
              RowBox[{
               SuperscriptBox["gg", "\[Prime]",
                MultilineFunction->None], "[", "v", "]"}]}]}], ")"}]}]}], 
         ")"}]}]}], ")"}]}], "+", 
    RowBox[{"u", " ", 
     RowBox[{"Cos", "[", 
      RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
     RowBox[{
      SuperscriptBox["gg", "\[Prime]",
       MultilineFunction->None], "[", "v", "]"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "u"}], " ", 
           RowBox[{"Cos", "[", 
            RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
           RowBox[{"Sin", "[", 
            RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
           RowBox[{
            SuperscriptBox["ff", "\[Prime]",
             MultilineFunction->None], "[", "v", "]"}]}], "-", 
          RowBox[{"u", " ", 
           RowBox[{"Cos", "[", 
            RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
           RowBox[{"Sin", "[", 
            RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
           RowBox[{
            SuperscriptBox["gg", "\[Prime]",
             MultilineFunction->None], "[", "v", "]"}]}]}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"2", " ", 
           SuperscriptBox[
            RowBox[{"Cos", "[", 
             RowBox[{"gg", "[", "v", "]"}], "]"}], "3"], " ", 
           SuperscriptBox[
            RowBox[{"Sin", "[", 
             RowBox[{"ff", "[", "v", "]"}], "]"}], "2"], " ", 
           RowBox[{
            SuperscriptBox["gg", "\[Prime]",
             MultilineFunction->None], "[", "v", "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", 
               RowBox[{"Cos", "[", 
                RowBox[{"gg", "[", "v", "]"}], "]"}]}], " ", 
              RowBox[{"Sin", "[", 
               RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
              RowBox[{
               SuperscriptBox["ff", "\[Prime]",
                MultilineFunction->None], "[", "v", "]"}]}], "-", 
             RowBox[{
              RowBox[{"Cos", "[", 
               RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
              RowBox[{"Sin", "[", 
               RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
              RowBox[{
               SuperscriptBox["gg", "\[Prime]",
                MultilineFunction->None], "[", "v", "]"}]}]}], ")"}]}], "+", 
          RowBox[{"2", " ", 
           RowBox[{"Cos", "[", 
            RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
           RowBox[{"Cos", "[", 
            RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
           RowBox[{"Sin", "[", 
            RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"Cos", "[", 
                RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
               RowBox[{"Cos", "[", 
                RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
               RowBox[{
                SuperscriptBox["ff", "\[Prime]",
                 MultilineFunction->None], "[", "v", "]"}]}], "-", 
              RowBox[{
               RowBox[{"Sin", "[", 
                RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
               RowBox[{"Sin", "[", 
                RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
               RowBox[{
                SuperscriptBox["gg", "\[Prime]",
                 MultilineFunction->None], "[", "v", "]"}]}]}], ")"}], "2"]}],
           "+", 
          RowBox[{
           RowBox[{"Cos", "[", 
            RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
           RowBox[{"Sin", "[", 
            RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "2"}], " ", 
              RowBox[{"Cos", "[", 
               RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
              SuperscriptBox[
               RowBox[{"Cos", "[", 
                RowBox[{"gg", "[", "v", "]"}], "]"}], "2"], " ", 
              RowBox[{
               SuperscriptBox["gg", "\[Prime]",
                MultilineFunction->None], "[", "v", "]"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"Cos", "[", 
                  RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
                 RowBox[{"Cos", "[", 
                  RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
                 RowBox[{
                  SuperscriptBox["ff", "\[Prime]",
                   MultilineFunction->None], "[", "v", "]"}]}], "-", 
                RowBox[{
                 RowBox[{"Sin", "[", 
                  RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
                 RowBox[{"Sin", "[", 
                  RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
                 RowBox[{
                  SuperscriptBox["gg", "\[Prime]",
                   MultilineFunction->None], "[", "v", "]"}]}]}], ")"}]}], 
             "-", 
             RowBox[{"2", " ", 
              RowBox[{"Sin", "[", 
               RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", 
                  RowBox[{"Cos", "[", 
                   RowBox[{"gg", "[", "v", "]"}], "]"}]}], " ", 
                 RowBox[{"Sin", "[", 
                  RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
                 RowBox[{
                  SuperscriptBox["ff", "\[Prime]",
                   MultilineFunction->None], "[", "v", "]"}]}], "-", 
                RowBox[{
                 RowBox[{"Cos", "[", 
                  RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
                 RowBox[{"Sin", "[", 
                  RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
                 RowBox[{
                  SuperscriptBox["gg", "\[Prime]",
                   MultilineFunction->None], "[", "v", "]"}]}]}], ")"}], " ", 
              
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"Cos", "[", 
                  RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
                 RowBox[{"Cos", "[", 
                  RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
                 RowBox[{
                  SuperscriptBox["ff", "\[Prime]",
                   MultilineFunction->None], "[", "v", "]"}]}], "-", 
                RowBox[{
                 RowBox[{"Sin", "[", 
                  RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
                 RowBox[{"Sin", "[", 
                  RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
                 RowBox[{
                  SuperscriptBox["gg", "\[Prime]",
                   MultilineFunction->None], "[", "v", "]"}]}]}], ")"}]}]}], 
            ")"}]}]}], ")"}]}], "+", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"u", " ", 
           RowBox[{"Cos", "[", 
            RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
           RowBox[{"Cos", "[", 
            RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
           RowBox[{
            SuperscriptBox["ff", "\[Prime]",
             MultilineFunction->None], "[", "v", "]"}]}], "-", 
          RowBox[{"u", " ", 
           RowBox[{"Sin", "[", 
            RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
           RowBox[{"Sin", "[", 
            RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
           RowBox[{
            SuperscriptBox["gg", "\[Prime]",
             MultilineFunction->None], "[", "v", "]"}]}]}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"Cos", "[", 
            RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
           RowBox[{"Sin", "[", 
            RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
              RowBox[{"-", "2"}], " ", 
              RowBox[{"Cos", "[", 
               RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
              SuperscriptBox[
               RowBox[{"Cos", "[", 
                RowBox[{"gg", "[", "v", "]"}], "]"}], "2"], " ", 
              RowBox[{
               SuperscriptBox["gg", "\[Prime]",
                MultilineFunction->None], "[", "v", "]"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", 
                  RowBox[{"Cos", "[", 
                   RowBox[{"gg", "[", "v", "]"}], "]"}]}], " ", 
                 RowBox[{"Sin", "[", 
                  RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
                 RowBox[{
                  SuperscriptBox["ff", "\[Prime]",
                   MultilineFunction->None], "[", "v", "]"}]}], "-", 
                RowBox[{
                 RowBox[{"Cos", "[", 
                  RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
                 RowBox[{"Sin", "[", 
                  RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
                 RowBox[{
                  SuperscriptBox["gg", "\[Prime]",
                   MultilineFunction->None], "[", "v", "]"}]}]}], ")"}]}], 
             "+", 
             RowBox[{"2", " ", 
              RowBox[{"Sin", "[", 
               RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{
                 RowBox[{
                  RowBox[{"-", 
                   RowBox[{"Cos", "[", 
                    RowBox[{"gg", "[", "v", "]"}], "]"}]}], " ", 
                  RowBox[{"Sin", "[", 
                   RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox["ff", "\[Prime]",
                    MultilineFunction->None], "[", "v", "]"}]}], "-", 
                 RowBox[{
                  RowBox[{"Cos", "[", 
                   RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
                  RowBox[{"Sin", "[", 
                   RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
                  RowBox[{
                   SuperscriptBox["gg", "\[Prime]",
                    MultilineFunction->None], "[", "v", "]"}]}]}], ")"}], 
               "2"]}]}], ")"}]}], "+", 
          RowBox[{
           RowBox[{"Cos", "[", 
            RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
           RowBox[{"Cos", "[", 
            RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"2", " ", 
              RowBox[{"Cos", "[", 
               RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
              SuperscriptBox[
               RowBox[{"Cos", "[", 
                RowBox[{"gg", "[", "v", "]"}], "]"}], "2"], " ", 
              RowBox[{
               SuperscriptBox["gg", "\[Prime]",
                MultilineFunction->None], "[", "v", "]"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"Cos", "[", 
                  RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
                 RowBox[{"Cos", "[", 
                  RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
                 RowBox[{
                  SuperscriptBox["ff", "\[Prime]",
                   MultilineFunction->None], "[", "v", "]"}]}], "-", 
                RowBox[{
                 RowBox[{"Sin", "[", 
                  RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
                 RowBox[{"Sin", "[", 
                  RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
                 RowBox[{
                  SuperscriptBox["gg", "\[Prime]",
                   MultilineFunction->None], "[", "v", "]"}]}]}], ")"}]}], 
             "-", 
             RowBox[{"2", " ", 
              RowBox[{"Sin", "[", 
               RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"-", 
                  RowBox[{"Cos", "[", 
                   RowBox[{"gg", "[", "v", "]"}], "]"}]}], " ", 
                 RowBox[{"Sin", "[", 
                  RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
                 RowBox[{
                  SuperscriptBox["ff", "\[Prime]",
                   MultilineFunction->None], "[", "v", "]"}]}], "-", 
                RowBox[{
                 RowBox[{"Cos", "[", 
                  RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
                 RowBox[{"Sin", "[", 
                  RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
                 RowBox[{
                  SuperscriptBox["gg", "\[Prime]",
                   MultilineFunction->None], "[", "v", "]"}]}]}], ")"}], " ", 
              
              RowBox[{"(", 
               RowBox[{
                RowBox[{
                 RowBox[{"Cos", "[", 
                  RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
                 RowBox[{"Cos", "[", 
                  RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
                 RowBox[{
                  SuperscriptBox["ff", "\[Prime]",
                   MultilineFunction->None], "[", "v", "]"}]}], "-", 
                RowBox[{
                 RowBox[{"Sin", "[", 
                  RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
                 RowBox[{"Sin", "[", 
                  RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
                 RowBox[{
                  SuperscriptBox["gg", "\[Prime]",
                   MultilineFunction->None], "[", "v", "]"}]}]}], ")"}]}]}], 
            ")"}]}]}], ")"}]}]}], ")"}]}]}], ")"}], "/", 
  SuperscriptBox[
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      SuperscriptBox["u", "2"], " ", 
      SuperscriptBox[
       RowBox[{"Cos", "[", 
        RowBox[{"gg", "[", "v", "]"}], "]"}], "2"], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         SuperscriptBox[
          RowBox[{"Cos", "[", 
           RowBox[{"ff", "[", "v", "]"}], "]"}], "2"], " ", 
         SuperscriptBox[
          RowBox[{"Cos", "[", 
           RowBox[{"gg", "[", "v", "]"}], "]"}], "2"]}], "+", 
        RowBox[{
         SuperscriptBox[
          RowBox[{"Cos", "[", 
           RowBox[{"gg", "[", "v", "]"}], "]"}], "2"], " ", 
         SuperscriptBox[
          RowBox[{"Sin", "[", 
           RowBox[{"ff", "[", "v", "]"}], "]"}], "2"]}]}], ")"}], " ", 
      SuperscriptBox[
       RowBox[{
        SuperscriptBox["gg", "\[Prime]",
         MultilineFunction->None], "[", "v", "]"}], "2"]}], "-", 
     RowBox[{"2", " ", "u", " ", 
      RowBox[{"Cos", "[", 
       RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
      SuperscriptBox[
       RowBox[{"Cos", "[", 
        RowBox[{"gg", "[", "v", "]"}], "]"}], "2"], " ", 
      RowBox[{"Sin", "[", 
       RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
      RowBox[{
       SuperscriptBox["gg", "\[Prime]",
        MultilineFunction->None], "[", "v", "]"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "u"}], " ", 
         RowBox[{"Cos", "[", 
          RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
         RowBox[{"Sin", "[", 
          RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
         RowBox[{
          SuperscriptBox["ff", "\[Prime]",
           MultilineFunction->None], "[", "v", "]"}]}], "-", 
        RowBox[{"u", " ", 
         RowBox[{"Cos", "[", 
          RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
         RowBox[{"Sin", "[", 
          RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
         RowBox[{
          SuperscriptBox["gg", "\[Prime]",
           MultilineFunction->None], "[", "v", "]"}]}]}], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         SuperscriptBox[
          RowBox[{"Cos", "[", 
           RowBox[{"gg", "[", "v", "]"}], "]"}], "2"], " ", 
         SuperscriptBox[
          RowBox[{"Sin", "[", 
           RowBox[{"ff", "[", "v", "]"}], "]"}], "2"]}], "+", 
        SuperscriptBox[
         RowBox[{"Sin", "[", 
          RowBox[{"gg", "[", "v", "]"}], "]"}], "2"]}], ")"}], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", "u"}], " ", 
          RowBox[{"Cos", "[", 
           RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
          RowBox[{"Sin", "[", 
           RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
          RowBox[{
           SuperscriptBox["ff", "\[Prime]",
            MultilineFunction->None], "[", "v", "]"}]}], "-", 
         RowBox[{"u", " ", 
          RowBox[{"Cos", "[", 
           RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
          RowBox[{"Sin", "[", 
           RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
          RowBox[{
           SuperscriptBox["gg", "\[Prime]",
            MultilineFunction->None], "[", "v", "]"}]}]}], ")"}], "2"]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         SuperscriptBox[
          RowBox[{"Cos", "[", 
           RowBox[{"ff", "[", "v", "]"}], "]"}], "2"], " ", 
         SuperscriptBox[
          RowBox[{"Cos", "[", 
           RowBox[{"gg", "[", "v", "]"}], "]"}], "2"]}], "+", 
        SuperscriptBox[
         RowBox[{"Sin", "[", 
          RowBox[{"gg", "[", "v", "]"}], "]"}], "2"]}], ")"}], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         RowBox[{"u", " ", 
          RowBox[{"Cos", "[", 
           RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
          RowBox[{"Cos", "[", 
           RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
          RowBox[{
           SuperscriptBox["ff", "\[Prime]",
            MultilineFunction->None], "[", "v", "]"}]}], "-", 
         RowBox[{"u", " ", 
          RowBox[{"Sin", "[", 
           RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
          RowBox[{"Sin", "[", 
           RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
          RowBox[{
           SuperscriptBox["gg", "\[Prime]",
            MultilineFunction->None], "[", "v", "]"}]}]}], ")"}], "2"]}], "-", 
     RowBox[{"2", " ", 
      RowBox[{"Cos", "[", 
       RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
      RowBox[{"Sin", "[", 
       RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"u", " ", 
         RowBox[{"Cos", "[", 
          RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
         RowBox[{"Cos", "[", 
          RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
         RowBox[{
          SuperscriptBox["ff", "\[Prime]",
           MultilineFunction->None], "[", "v", "]"}]}], "-", 
        RowBox[{"u", " ", 
         RowBox[{"Sin", "[", 
          RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
         RowBox[{"Sin", "[", 
          RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
         RowBox[{
          SuperscriptBox["gg", "\[Prime]",
           MultilineFunction->None], "[", "v", "]"}]}]}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"u", " ", 
         RowBox[{"Cos", "[", 
          RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
         RowBox[{"Sin", "[", 
          RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
         RowBox[{
          SuperscriptBox["gg", "\[Prime]",
           MultilineFunction->None], "[", "v", "]"}]}], "+", 
        RowBox[{
         RowBox[{"Cos", "[", 
          RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
         RowBox[{"Cos", "[", 
          RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", "u"}], " ", 
            RowBox[{"Cos", "[", 
             RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
            RowBox[{"Sin", "[", 
             RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
            RowBox[{
             SuperscriptBox["ff", "\[Prime]",
              MultilineFunction->None], "[", "v", "]"}]}], "-", 
           RowBox[{"u", " ", 
            RowBox[{"Cos", "[", 
             RowBox[{"ff", "[", "v", "]"}], "]"}], " ", 
            RowBox[{"Sin", "[", 
             RowBox[{"gg", "[", "v", "]"}], "]"}], " ", 
            RowBox[{
             SuperscriptBox["gg", "\[Prime]",
              MultilineFunction->None], "[", "v", "]"}]}]}], ")"}]}]}], 
       ")"}]}]}], ")"}], "2"]}]], "Output",
 CellChangeTimes->{
  3.707566996617004*^9},ExpressionUUID->"1320a40b-b05e-4e7e-a013-\
6eddae96f6e3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FullSimplify", "[", 
  RowBox[{"%", ",", " ", 
   RowBox[{"v", ">", "0"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.7075670083424187`*^9, 
  3.707567041338336*^9}},ExpressionUUID->"d623b481-c26e-422c-8713-\
70f059abaf92"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{
  3.707567043124095*^9},ExpressionUUID->"8272bad4-8f8e-49de-8d60-\
876704fbe465"]
}, Open  ]],

Cell["\<\
\[FilledSquare]
A tangent surface of a helix\
\>", "Text",
 CellChangeTimes->{{3.70757182809803*^9, 
  3.707571849811631*^9}},ExpressionUUID->"808c9967-c62a-4d30-b709-\
5fcd58626c7b"],

Cell[BoxData[
 RowBox[{
  RowBox[{"Clear", "[", "tang", "]"}], ";", 
  RowBox[{
   RowBox[{"tang", "[", 
    RowBox[{"u_", ",", "v_"}], "]"}], " ", ":=", " ", 
   RowBox[{
    RowBox[{
     RowBox[{"helix", "[", 
      RowBox[{"1", ",", 
       RowBox[{"1", "/", "3"}]}], "]"}], "[", "v", "]"}], "+", " ", 
    RowBox[{"u", "*", 
     RowBox[{
      RowBox[{
       RowBox[{"helix", "[", 
        RowBox[{"1", ",", 
         RowBox[{"1", "/", "3"}]}], "]"}], "'"}], "[", "v", 
      "]"}]}]}]}]}]], "Input",
 CellChangeTimes->{{3.707567102990821*^9, 
  3.707567232761681*^9}},ExpressionUUID->"fdcfcb46-e83f-4a18-88bf-\
92255a2aebd9"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"tang", "[", 
  RowBox[{"u", ",", "v"}], "]"}]], "Input",
 CellChangeTimes->{{3.7075672432711906`*^9, 
  3.707567253871586*^9}},ExpressionUUID->"6ff47261-b161-4e8b-9edb-\
544ab45046d9"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"Cos", "[", "v", "]"}], "-", 
    RowBox[{"u", " ", 
     RowBox[{"Sin", "[", "v", "]"}]}]}], ",", 
   RowBox[{
    RowBox[{"u", " ", 
     RowBox[{"Cos", "[", "v", "]"}]}], "+", 
    RowBox[{"Sin", "[", "v", "]"}]}], ",", 
   RowBox[{
    FractionBox["u", "3"], "+", 
    FractionBox["v", "3"]}]}], "}"}]], "Output",
 CellChangeTimes->{
  3.707567255130887*^9},ExpressionUUID->"7460d1eb-028e-4cd1-8644-\
bdcb22b2c233"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"gausscurvature", "[", "tang", "]"}], "[", 
  RowBox[{"u", ",", "v"}], "]"}]], "Input",
 CellChangeTimes->{{3.707567264652858*^9, 
  3.707567293192926*^9}},ExpressionUUID->"6cbfaa6e-f14f-45d4-8b6f-\
4ac0d0f3886c"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      FractionBox["1", "9"]}], " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "u"}], " ", 
         RowBox[{"Cos", "[", "v", "]"}]}], "-", 
        RowBox[{"Sin", "[", "v", "]"}]}], ")"}], "2"], " ", 
     SuperscriptBox[
      RowBox[{"Sin", "[", "v", "]"}], "2"]}], "+", 
    RowBox[{
     FractionBox["2", "9"], " ", 
     RowBox[{"Cos", "[", "v", "]"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "u"}], " ", 
        RowBox[{"Cos", "[", "v", "]"}]}], "-", 
       RowBox[{"Sin", "[", "v", "]"}]}], ")"}], " ", 
     RowBox[{"Sin", "[", "v", "]"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"Cos", "[", "v", "]"}], "-", 
       RowBox[{"u", " ", 
        RowBox[{"Sin", "[", "v", "]"}]}]}], ")"}]}], "-", 
    RowBox[{
     FractionBox["1", "9"], " ", 
     SuperscriptBox[
      RowBox[{"Cos", "[", "v", "]"}], "2"], " ", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{
        RowBox[{"Cos", "[", "v", "]"}], "-", 
        RowBox[{"u", " ", 
         RowBox[{"Sin", "[", "v", "]"}]}]}], ")"}], "2"]}], "+", 
    RowBox[{
     FractionBox["1", "9"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", 
        SuperscriptBox[
         RowBox[{"Cos", "[", "v", "]"}], "4"]}], "-", 
       RowBox[{"2", " ", 
        SuperscriptBox[
         RowBox[{"Cos", "[", "v", "]"}], "2"], " ", 
        SuperscriptBox[
         RowBox[{"Sin", "[", "v", "]"}], "2"]}], "-", 
       SuperscriptBox[
        RowBox[{"Sin", "[", "v", "]"}], "4"]}], ")"}]}], "+", 
    RowBox[{
     FractionBox["1", "3"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"Cos", "[", "v", "]"}], "-", 
          RowBox[{"u", " ", 
           RowBox[{"Sin", "[", "v", "]"}]}]}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          FractionBox[
           RowBox[{"2", " ", 
            SuperscriptBox[
             RowBox[{"Cos", "[", "v", "]"}], "3"]}], "3"], "+", 
          RowBox[{
           FractionBox["2", "3"], " ", 
           RowBox[{"Cos", "[", "v", "]"}], " ", 
           SuperscriptBox[
            RowBox[{"Sin", "[", "v", "]"}], "2"]}]}], ")"}]}], "+", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "u"}], " ", 
           RowBox[{"Cos", "[", "v", "]"}]}], "-", 
          RowBox[{"Sin", "[", "v", "]"}]}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", 
            FractionBox["2", "3"]}], " ", 
           SuperscriptBox[
            RowBox[{"Cos", "[", "v", "]"}], "2"], " ", 
           RowBox[{"Sin", "[", "v", "]"}]}], "-", 
          FractionBox[
           RowBox[{"2", " ", 
            SuperscriptBox[
             RowBox[{"Sin", "[", "v", "]"}], "3"]}], "3"]}], ")"}]}]}], 
      ")"}]}]}], ")"}], "/", 
  SuperscriptBox[
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        FractionBox["1", "9"], "+", 
        SuperscriptBox[
         RowBox[{"Cos", "[", "v", "]"}], "2"]}], ")"}], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", "u"}], " ", 
          RowBox[{"Cos", "[", "v", "]"}]}], "-", 
         RowBox[{"Sin", "[", "v", "]"}]}], ")"}], "2"]}], "+", 
     RowBox[{
      FractionBox["2", "9"], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "u"}], " ", 
         RowBox[{"Cos", "[", "v", "]"}]}], "-", 
        RowBox[{"Sin", "[", "v", "]"}]}], ")"}], " ", 
      RowBox[{"Sin", "[", "v", "]"}]}], "-", 
     RowBox[{"2", " ", 
      RowBox[{"Cos", "[", "v", "]"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"Cos", "[", "v", "]"}], "-", 
        RowBox[{"u", " ", 
         RowBox[{"Sin", "[", "v", "]"}]}]}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        FractionBox["1", "9"], "-", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", "u"}], " ", 
            RowBox[{"Cos", "[", "v", "]"}]}], "-", 
           RowBox[{"Sin", "[", "v", "]"}]}], ")"}], " ", 
         RowBox[{"Sin", "[", "v", "]"}]}]}], ")"}]}], "+", 
     RowBox[{
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         RowBox[{"Cos", "[", "v", "]"}], "-", 
         RowBox[{"u", " ", 
          RowBox[{"Sin", "[", "v", "]"}]}]}], ")"}], "2"], " ", 
      RowBox[{"(", 
       RowBox[{
        FractionBox["1", "9"], "+", 
        SuperscriptBox[
         RowBox[{"Sin", "[", "v", "]"}], "2"]}], ")"}]}], "+", 
     RowBox[{
      FractionBox["1", "9"], " ", 
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox[
         RowBox[{"Cos", "[", "v", "]"}], "2"], "+", 
        SuperscriptBox[
         RowBox[{"Sin", "[", "v", "]"}], "2"]}], ")"}]}]}], ")"}], 
   "2"]}]], "Output",
 CellChangeTimes->{
  3.707567300927252*^9},ExpressionUUID->"f65ae86b-47c0-4141-a330-\
837b73f2fa2f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FullSimplify", "[", 
  RowBox[{"%", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"u", ">", "0"}], ",", 
     RowBox[{"v", ">", "0"}]}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.707567311053893*^9, 
  3.7075673511201477`*^9}},ExpressionUUID->"454b7483-6b87-4e54-9358-\
0140203bb7de"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{
  3.7075673544340277`*^9},ExpressionUUID->"a8cfd252-54ba-4218-b945-\
b43251753e20"]
}, Open  ]],

Cell["\<\
\[FilledSquare]
A general hyperboloid\
\>", "Text",
 CellChangeTimes->{{3.7075718714374313`*^9, 
  3.707571889108552*^9}},ExpressionUUID->"fd306329-504f-4a61-b044-\
cb31cf26760e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"hyperboloid", "[", 
   RowBox[{"a", ",", "b", ",", "c"}], "]"}], "[", 
  RowBox[{"u", ",", "v"}], "]"}]], "Input",
 CellChangeTimes->{
  3.707630163926819*^9},ExpressionUUID->"72bc9c41-81e6-40ca-9e12-\
13717653be1c"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"a", " ", 
    RowBox[{"Cos", "[", "u", "]"}], " ", 
    RowBox[{"Cosh", "[", "v", "]"}]}], ",", 
   RowBox[{"b", " ", 
    RowBox[{"Cosh", "[", "v", "]"}], " ", 
    RowBox[{"Sin", "[", "u", "]"}]}], ",", 
   RowBox[{"c", " ", 
    RowBox[{"Sinh", "[", "v", "]"}]}]}], "}"}]], "Output",
 CellChangeTimes->{
  3.707630166529264*^9},ExpressionUUID->"96e9451c-b277-40a3-a160-\
d5a8c6f93cf8"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"gausscurvature", "[", 
   RowBox[{"hyperboloid", "[", 
    RowBox[{"a", ",", "b", ",", "c"}], "]"}], "]"}], "[", 
  RowBox[{"u", ",", "v"}], "]"}]], "Input",
 CellChangeTimes->{{3.707567492080359*^9, 
  3.707567530415663*^9}},ExpressionUUID->"6ada5c36-d66d-4880-a70e-\
3ec2d60f8d6a"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{
    RowBox[{"c", " ", 
     RowBox[{"Cosh", "[", "v", "]"}], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"a", " ", 
        RowBox[{"Cos", "[", "u", "]"}], " ", 
        RowBox[{"Sinh", "[", "v", "]"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"a", " ", 
           SuperscriptBox["b", "2"], " ", "c", " ", 
           SuperscriptBox[
            RowBox[{"Cos", "[", "u", "]"}], "3"], " ", 
           SuperscriptBox[
            RowBox[{"Cosh", "[", "v", "]"}], "3"], " ", 
           RowBox[{"Sinh", "[", "v", "]"}]}], "+", 
          RowBox[{"a", " ", 
           SuperscriptBox["b", "2"], " ", "c", " ", 
           RowBox[{"Cos", "[", "u", "]"}], " ", 
           SuperscriptBox[
            RowBox[{"Cosh", "[", "v", "]"}], "3"], " ", 
           SuperscriptBox[
            RowBox[{"Sin", "[", "u", "]"}], "2"], " ", 
           RowBox[{"Sinh", "[", "v", "]"}]}]}], ")"}]}], "+", 
       RowBox[{"b", " ", 
        RowBox[{"Sin", "[", "u", "]"}], " ", 
        RowBox[{"Sinh", "[", "v", "]"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           SuperscriptBox["a", "2"], " ", "b", " ", "c", " ", 
           SuperscriptBox[
            RowBox[{"Cos", "[", "u", "]"}], "2"], " ", 
           SuperscriptBox[
            RowBox[{"Cosh", "[", "v", "]"}], "3"], " ", 
           RowBox[{"Sin", "[", "u", "]"}], " ", 
           RowBox[{"Sinh", "[", "v", "]"}]}], "+", 
          RowBox[{
           SuperscriptBox["a", "2"], " ", "b", " ", "c", " ", 
           SuperscriptBox[
            RowBox[{"Cosh", "[", "v", "]"}], "3"], " ", 
           SuperscriptBox[
            RowBox[{"Sin", "[", "u", "]"}], "3"], " ", 
           RowBox[{"Sinh", "[", "v", "]"}]}]}], ")"}]}]}], ")"}]}], "-", 
    RowBox[{
     SuperscriptBox["c", "2"], " ", 
     SuperscriptBox[
      RowBox[{"Cosh", "[", "v", "]"}], "2"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        SuperscriptBox["a", "2"], " ", 
        SuperscriptBox[
         RowBox[{"Cosh", "[", "v", "]"}], "2"], " ", 
        SuperscriptBox[
         RowBox[{"Sin", "[", "u", "]"}], "2"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           SuperscriptBox["b", "2"], " ", 
           SuperscriptBox[
            RowBox[{"Cosh", "[", "v", "]"}], "2"], " ", 
           SuperscriptBox[
            RowBox[{"Sin", "[", "u", "]"}], "2"]}], "+", 
          RowBox[{
           SuperscriptBox["b", "2"], " ", 
           SuperscriptBox[
            RowBox[{"Cos", "[", "u", "]"}], "2"], " ", 
           SuperscriptBox[
            RowBox[{"Sinh", "[", "v", "]"}], "2"]}]}], ")"}]}], "-", 
       RowBox[{"a", " ", "b", " ", 
        RowBox[{"Cos", "[", "u", "]"}], " ", 
        SuperscriptBox[
         RowBox[{"Cosh", "[", "v", "]"}], "2"], " ", 
        RowBox[{"Sin", "[", "u", "]"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "2"}], " ", "a", " ", "b", " ", 
           RowBox[{"Cos", "[", "u", "]"}], " ", 
           SuperscriptBox[
            RowBox[{"Cosh", "[", "v", "]"}], "2"], " ", 
           RowBox[{"Sin", "[", "u", "]"}]}], "+", 
          RowBox[{"2", " ", "a", " ", "b", " ", 
           RowBox[{"Cos", "[", "u", "]"}], " ", 
           RowBox[{"Sin", "[", "u", "]"}], " ", 
           SuperscriptBox[
            RowBox[{"Sinh", "[", "v", "]"}], "2"]}]}], ")"}]}], "+", 
       RowBox[{
        SuperscriptBox["b", "2"], " ", 
        SuperscriptBox[
         RowBox[{"Cos", "[", "u", "]"}], "2"], " ", 
        SuperscriptBox[
         RowBox[{"Cosh", "[", "v", "]"}], "2"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           SuperscriptBox["a", "2"], " ", 
           SuperscriptBox[
            RowBox[{"Cos", "[", "u", "]"}], "2"], " ", 
           SuperscriptBox[
            RowBox[{"Cosh", "[", "v", "]"}], "2"]}], "+", 
          RowBox[{
           SuperscriptBox["a", "2"], " ", 
           SuperscriptBox[
            RowBox[{"Sin", "[", "u", "]"}], "2"], " ", 
           SuperscriptBox[
            RowBox[{"Sinh", "[", "v", "]"}], "2"]}]}], ")"}]}]}], ")"}]}]}], 
   ")"}], "/", 
  SuperscriptBox[
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      SuperscriptBox["c", "2"], " ", 
      SuperscriptBox[
       RowBox[{"Cosh", "[", "v", "]"}], "2"], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         SuperscriptBox["b", "2"], " ", 
         SuperscriptBox[
          RowBox[{"Cos", "[", "u", "]"}], "2"], " ", 
         SuperscriptBox[
          RowBox[{"Cosh", "[", "v", "]"}], "2"]}], "+", 
        RowBox[{
         SuperscriptBox["a", "2"], " ", 
         SuperscriptBox[
          RowBox[{"Cosh", "[", "v", "]"}], "2"], " ", 
         SuperscriptBox[
          RowBox[{"Sin", "[", "u", "]"}], "2"]}]}], ")"}]}], "+", 
     RowBox[{
      SuperscriptBox["a", "2"], " ", 
      SuperscriptBox["b", "2"], " ", 
      SuperscriptBox[
       RowBox[{"Cos", "[", "u", "]"}], "4"], " ", 
      SuperscriptBox[
       RowBox[{"Cosh", "[", "v", "]"}], "2"], " ", 
      SuperscriptBox[
       RowBox[{"Sinh", "[", "v", "]"}], "2"]}], "+", 
     RowBox[{"2", " ", 
      SuperscriptBox["a", "2"], " ", 
      SuperscriptBox["b", "2"], " ", 
      SuperscriptBox[
       RowBox[{"Cos", "[", "u", "]"}], "2"], " ", 
      SuperscriptBox[
       RowBox[{"Cosh", "[", "v", "]"}], "2"], " ", 
      SuperscriptBox[
       RowBox[{"Sin", "[", "u", "]"}], "2"], " ", 
      SuperscriptBox[
       RowBox[{"Sinh", "[", "v", "]"}], "2"]}], "+", 
     RowBox[{
      SuperscriptBox["a", "2"], " ", 
      SuperscriptBox["b", "2"], " ", 
      SuperscriptBox[
       RowBox[{"Cosh", "[", "v", "]"}], "2"], " ", 
      SuperscriptBox[
       RowBox[{"Sin", "[", "u", "]"}], "4"], " ", 
      SuperscriptBox[
       RowBox[{"Sinh", "[", "v", "]"}], "2"]}]}], ")"}], "2"]}]], "Output",
 CellChangeTimes->{3.707567542657999*^9, 
  3.707568005790903*^9},ExpressionUUID->"2b24126e-98b4-4deb-a2bd-\
435ca33f5fcd"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FullSimplify", "[", "%", "]"}]], "Input",
 CellChangeTimes->{{3.707567572933989*^9, 
  3.7075675873270617`*^9}},ExpressionUUID->"7f565a74-40e8-479b-968d-\
3260a891df7e"],

Cell[BoxData[
 RowBox[{"-", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{"16", " ", 
     SuperscriptBox["a", "2"], " ", 
     SuperscriptBox["b", "2"], " ", 
     SuperscriptBox["c", "2"]}], ")"}], "/", 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "2"}], " ", 
       SuperscriptBox["a", "2"], " ", 
       SuperscriptBox["b", "2"]}], "+", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         SuperscriptBox["a", "2"], "+", 
         SuperscriptBox["b", "2"]}], ")"}], " ", 
       SuperscriptBox["c", "2"]}], "+", 
      RowBox[{"2", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", 
          SuperscriptBox["a", "2"]}], "+", 
         SuperscriptBox["b", "2"]}], ")"}], " ", 
       SuperscriptBox["c", "2"], " ", 
       RowBox[{"Cos", "[", 
        RowBox[{"2", " ", "u"}], "]"}], " ", 
       SuperscriptBox[
        RowBox[{"Cosh", "[", "v", "]"}], "2"]}], "+", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"2", " ", 
          SuperscriptBox["a", "2"], " ", 
          SuperscriptBox["b", "2"]}], "+", 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            SuperscriptBox["a", "2"], "+", 
            SuperscriptBox["b", "2"]}], ")"}], " ", 
          SuperscriptBox["c", "2"]}]}], ")"}], " ", 
       RowBox[{"Cosh", "[", 
        RowBox[{"2", " ", "v"}], "]"}]}]}], ")"}], "2"]}]}]], "Output",
 CellChangeTimes->{3.707567591114006*^9, 
  3.7075680163718853`*^9},ExpressionUUID->"b8876d13-97f8-45d2-a1e0-\
827336bceae5"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Plot3D", "[", 
  RowBox[{
   RowBox[{"%", "/.", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"a", "\[Rule]", "1"}], ",", 
      RowBox[{"b", "\[Rule]", 
       RowBox[{"1", "/", "3"}]}], ",", 
      RowBox[{"c", "\[Rule]", "2"}]}], "}"}]}], ",", 
   RowBox[{"{", 
    RowBox[{"u", ",", "0", ",", 
     RowBox[{"2", "*", "Pi"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"v", ",", 
     RowBox[{"-", "2"}], ",", "2"}], "}"}], ",", 
   RowBox[{"PlotRange", " ", "\[Rule]", " ", "All"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.707567623972391*^9, 3.707567734559132*^9}, {
  3.707567977607193*^9, 
  3.7075679897930193`*^9}},ExpressionUUID->"af18cf9a-e8d7-4608-86a9-\
cd5a859f0502"],

Cell[BoxData[
 Graphics3DBox[GraphicsComplex3DBox[CompressedData["
1:eJyFnXV0FkkT7oM7i7uTZXHN4kvjFhwWW2xZJMDi7u7u7g5hCRacDkFD0CAB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   "], {{
     {EdgeForm[None], GraphicsGroup3DBox[{Polygon3DBox[CompressedData["
1:eJxFnHf8V9Mfx7933/shq5CIipZRimRnZK9SNmmJKCnxsxLZexMhFckImVkZ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         "]], Polygon3DBox[CompressedData["
1:eJxFnAf4l9P/xp/P59nfQmUnlJTsMgolVEQhlZUdkUgLLaMSIjNbGRlZkV0U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         "]], Polygon3DBox[CompressedData["
1:eJwl13m8T3Uex/Fffve6LrfFEtnGvieaipRs2RlyqyvTcDH2LamphxkzjyiN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         "]], 
        Polygon3DBox[{{1803, 1133, 1278, 2379, 1689, 1690}, {1796, 1130, 1272,
          2376, 1680, 1681}, {1692, 1691, 2380, 1279, 1280, 2381}, {1683, 
         1682, 2377, 1273, 1274, 2378}}]}]}, {}, {}, {}, {}}, {
     {GrayLevel[0], Line3DBox[CompressedData["
1:eJwl0j1I1HEcx/GvZk+a2S4VGT6fZy5OtrbYJA2SU0LgOSiIB3eLm9whYiFx
ReQTokJzDS41KFGTlPmspbS16NBa9Poj3N37/fn8f7/7/vjd3ekf7hkqi4g5
7/3yiMSfYS15cCnihvwIb2Odrp+nef5yREq+zweuRIzZ85AP8XHexzf43QsR
vdb91k3ZO6hv0mV0Z7p5XV5X4Eu8yCf4W/6Sr5lTsnaZf+Ez/D3/7Pl6MkO3
qPtrz6a8ix/k+oqI73IDbmEjbiezcQebcRdbcA9bcR9TeIBteIhpPMJ2/IH3
8Cd24HGy/6Jzm/fP3BN5Dz/Kv5zvm7yFa7ig+8Tf8dd8lS/xJ+4oZ0+JD/Mi
L/AdnnZPOWv/6HK6TV2j7qnuVDfnOzL6UT7JH/MRPsa7ead1Wd7Fs+6pWW5L
ZupuYUZ3U3dgT438FVfkeTOuy7NYjW/wWnJurMJXWIklvIov0NFiGn1lPEcj
YgpdT0xiRfK7olcUfZQnd+7hctn5/+4/5NZRRw==
       "]]}, {
      Line3DBox[{1606, 1896, 2414, 1764, 1605, 2415, 1765, 1607, 2416, 1766, 
       1608, 2417, 1767, 1609, 2418, 1768, 1610, 2419, 1769, 1611, 2420, 1770,
        1612, 1081, 1613, 2421, 1771, 1614, 2422, 1772, 1615, 2423, 1773, 
       1616, 2424, 1774, 1617, 2425, 1775, 1618, 1221, 1776, 1897}], 
      Line3DBox[{1620, 2289, 2290, 2329, 2604, 1619, 2291, 2292, 1088, 1621, 
       2426, 1777, 1622, 2427, 1778, 1623, 2428, 1779, 1624, 2429, 1780, 1625,
        2522, 1898, 1899, 1781, 1626, 2266, 2267, 2330, 2596, 1627, 2293, 
       2294, 1094, 1628, 2430, 1782, 1629, 2431, 1783, 1630, 2432, 1784, 1631,
        2433, 1785, 1632, 2523, 1900, 1901, 1786, 1633}], 
      Line3DBox[{1634, 1834, 2137, 2238, 2509, 2406, 2237, 2236, 2268, 2331, 
       2582, 1635, 1101, 1636, 2434, 1787, 1637, 2435, 1788, 1638, 2524, 1902,
        2220, 1789, 1903, 2525, 1835, 1905, 2161, 2162, 1904, 2163, 1836, 
       2139, 2241, 1528, 2240, 2239, 2272, 2332, 2583, 1639, 2333, 2461, 1640,
        2436, 1790, 1641, 2437, 1791, 1642, 2526, 1907, 2225, 1792, 1908, 
       2527, 1837, 1910, 1793, 1909}], Line3DBox[CompressedData["
1:eJwVzjlOgmEUheGLITJZamtiL/A7bcPCAgmhMg6F8iMoIWqhK3EHghHWgStw
QBeBiSb4ULzfuefcId/GUfsgzUTEPnrLnnzEXi6ij128FiIOMZaP8I0H+Qu9
p1O9FVzbLdEikmLEFf+oXkdXXZH92flFh1/iq+opKvjCpfzE/JrejttN2TY9
o8+Yq7fQRYKJ2RqGegN84E7+RG/pp14BbXfzNIeq2ylfz0a0aJn/MT/DBZ/h
y+o3bOId5/Jju6uLP7vbWPyZntIb+g9d8iZS
       "]], Line3DBox[CompressedData["
1:eJwNzVsygnEYB+A3NU6bkAtdyzmhkhbAEtzIqe6IDjNMLqxCsQZ1jwWwBbSI
zDCei2d+7++d7/1/84f1g1oiIvbpTEacT0dUpiJu2SOlp2YjVuUKZbsbdvmY
iWi7uTc/mSu09EeZpmmuyj6/bl/549p+JDPeLbk9NRflmRwy8n1SFuw67LCs
D1hiW2+zxYSe9E5WLpK3a7HJu3eu/OvO3OVLb+g9Ocel+Ug+MHb7wg8X9t9y
wbs5d8fmDXkin/n0fUKu2zVY402v8g8w2ygO
       "]], Line3DBox[CompressedData["
1:eJwVzjlSAlEUheGL2s3osAUIJXdATCGGJZiikDiEOIOKVS5CQfclhkCsiV8H
f52p3q1XPRl0+7mI6GCSRrTzES/YKURs468o65v8MZ5tV/RHf5ll25vuS27j
Sf6kNYz5UzrDBlIkGOmX3tZLEavslvdbdBOPtlRf5isY245oAw+2A3qIkb5E
i/h18962x+9nN2wX9Ft/Tu9sr7qp3MKt/EGruOF79B05rGMN1/qFt7v+cibP
vS/IeQxtiT7h0+y+rYU+/w9/wSP9
       "]], Line3DBox[CompressedData["
1:eJwVz8tKQlEYhuG/TLfuTtOySU7yOmoYROAVaBdQ04LSDqBZEHTAKLsoO15D
mZMKamjPHrx87/evxTpUtnZq2xMRsYmLQsRqMWINV0nErnwvRUzyPb6ARXzo
NfNf5PgzbtDWp+Slczo8z6/5Oj9FQX/BsjPGekUO9S8kGOh185k0Yp/3zA5k
UT7KJlq41Q+zd9hb4ke8jCV86t/4s5bKV9yhq0/Le+85y+7gD3yDn2NWf8OK
M8L9VTnSfzCHJ72R7bN2zPtmJ3I++7/8B3XeKH0=
       "]], Line3DBox[CompressedData["
1:eJwVzc0ygmEchvG/jcoqezKU04gZX303Y5s22bPnFBAqwpQVZvrgUAwnkQPQ
3s/imuu57/t53ne9dXp4shARdTwvRlSTETXcJCIqXEZqKWJoyzlv4seW5Xkq
YoMHtkfdh1zBk/zOOTw4l3mKvjt9+dWbZd98429diYu4tzXdS9oKctt2wPu4
s+3yHq70Ozxz9wg9W0Ze++9sq/yrX+GuraebyEXcymPO4tq5wCN03GnLL96k
/f9Y96XblrdwaWvoEra8fGE75zN84g/QcyeH
       "]], Line3DBox[CompressedData["
1:eJwVz8s2AlAUh/FtEhoIqcREA1aXd5GlN+gBeAl3Ss1FysDldRT1DowMmPmd
wbe+/97nnL3OrrSPD48WIqKJYYaXIg5wtxjxzLvLEevZiBe5ihq+nLX0f7Eq
f2CAG/UaP5rTSe/kkbwvd5FXf+LcjBUzL/hefcIb/M47+qd8hge9Sy7whK9w
jaH6lffSTPff5Doa+Hb2gz9nRZ6l/+BWXeKx//TkTfkp7Sv3UVbP0z0zcmb2
eJR24i2eckW/w12M9fq8nfbnf89HKaI=
       "]], 
      Line3DBox[{1701, 1864, 1957, 2037, 2038, 2594, 1956, 2194, 1865, 2196, 
       2402, 2514, 2195, 2011, 1958, 2450, 1820, 1702, 2451, 1821, 1703, 2409,
        2452, 1822, 2280, 2599, 1866, 2031, 2039, 2040, 1959, 2538, 2065, 
       1867, 1961, 2197, 2198, 1960, 2199, 1868, 1963, 2041, 2042, 2595, 1962,
        2200, 1869, 2203, 2201, 2515, 2202, 2012, 1964, 2453, 1823, 1704, 
       2454, 1824, 1705, 2410, 2455, 1825, 2281, 2600, 1870, 2032, 2043, 2044,
        1965, 2539, 2066, 1871, 1966, 1826, 2159}], Line3DBox[CompressedData["

1:eJwNzstKAlAUheEdhF2mQqSNIu3yLg3MSzWJUOflC/QeBd11lFoRVCSBTTQf
IHNU1LzeoKJv8LP+tc/ZhzNf2c3vjEXEKi4SEbXJiNxExDV+8TIVUcQVP3S2
Jt+xZDbEG79Bku+NRzS9sc8XzFp8djqiZy+j92VKz/MtfoRj3Oqbdn5Q4Pc4
MT/FHT+TRdnBH17dK+GBnzsryU8sm43wwR8xw1fwhba/HMis+SVP+8ez3UV9
IOf0db7N62igq3/bSfIN/oQKr6KMfwo6LTA=
       "]], Line3DBox[CompressedData["
1:eJwNzklOAlEUBdBnYlAhtgHLqSOMM5uFILagloYF6CrcBC5CkMYiBhPjJnBE
wkASWIPGMzh5v+599at2G4/Vh4WIqDDIRSTLEXdLEZ+kfK94JtNNdZfcc0Eh
H/Em37P3bKfv3DL36Tkf6jecE9bZoSvfMo91H96bMGTuvnMG1PV/3MrfuWFb
fkaJa88ZdUb2Ujru/dGdklIl7xuv8rK9pp2289OifzQPdGuyIquUeJFvmke6
rnfGdJi564SMmv6XmrzNFUX5l/kPZlspYA==
       "]], 
      Line3DBox[{1734, 1895, 2149, 1894, 2254, 2591, 1733, 2253, 2252, 2590, 
       1732, 1172, 1731, 2484, 2350, 1730, 2483, 2349, 1729, 2544, 1995, 2265,
        1994, 1728, 2547, 1893, 1988, 1892, 2136, 2235, 2234, 1891, 2148, 
       1890, 2251, 2589, 1727, 2250, 2249, 2588, 1726, 1167, 1725, 2482, 2348,
        1724, 2481, 2347, 1723, 2542, 1990, 2264, 1989, 1722, 2546, 1889, 
       1986, 1888, 1993}], 
      Line3DBox[{1736, 1991, 2019, 1992, 2543, 1735, 2351, 2485, 1737, 2352, 
       2486, 1738, 2353, 2487, 1739, 2354, 2488, 1740, 2355, 2489, 1741, 2601,
        2284, 2285, 2304, 2303, 1742, 1996, 2020, 1997, 2545, 1743, 2356, 
       2490, 1744, 2357, 2491, 1745, 2358, 2492, 1746, 2359, 2493, 1747, 2360,
        2494, 1748, 2460, 1833, 1749}], 
      Line3DBox[{1763, 2000, 2508, 2385, 1762, 2507, 2372, 1761, 2506, 2371, 
       1760, 2505, 2370, 1759, 2504, 2369, 1758, 2503, 2368, 1757, 2502, 2367,
        1756, 2501, 2366, 1755, 2500, 2365, 1754, 2499, 2364, 1753, 2498, 
       2363, 1752, 2497, 2362, 1751, 2496, 2361, 1750, 2384, 2495, 1998, 
       1999}]}, {Line3DBox[CompressedData["
1:eJwVz71KQgEYxvH3oAXl0hQSDkXUUCcojGgzIfqAWhy8gC4gobkLKCho6ga8
gIZQ0SlobrGk1mahRQwxoX4Nf97n6xzOWTqrVc6TiDjFVSaimI2ozURc09vT
EXt8nR5MRczPRpzwD/wdn/BVvs0v2j7SK7Kx7gcTPOvWZJv6pn7dTbGBF92c
7sOuwPf1KX2AQ3z7jnebWzqDLRzhGCPdp+6Cbnlux+3Kuu4bXtHX99y8d3ds
dumezSU91O3LF2RF31D6/1eU8atbln/Z3tjm0KDv3Tye6Lq7ij8OMCld
       "]], 
      Line3DBox[{906, 1076, 2416, 907, 1089, 2426, 932, 1101, 944, 1590, 2605,
        1111, 954, 2463, 1529, 1118, 964, 2468, 1578, 1124, 974, 2474, 1272, 
       1130, 2438, 984, 1284, 1136, 2444, 994, 1297, 1142, 2450, 1004, 1305, 
       1149, 1014, 1598, 2609, 1156, 1024, 2481, 1165, 1036, 2486, 1175, 1049,
        2497, 1188, 1063}], 
      Line3DBox[{908, 1077, 2417, 909, 1090, 2427, 933, 1102, 2434, 945, 1112,
        955, 2464, 1119, 965, 2469, 1125, 975, 2475, 1131, 2439, 985, 1137, 
       2445, 995, 1143, 2451, 1005, 1150, 2456, 1015, 1157, 1025, 2482, 1166, 
       1037, 2487, 1176, 1050, 2498, 1189, 1064}], 
      Line3DBox[{910, 1078, 2418, 911, 1091, 2428, 934, 1103, 2435, 946, 1592,
        2607, 1593, 956, 1245, 2531, 1246, 966, 2470, 1260, 1261, 976, 2476, 
       1273, 1274, 2440, 986, 1285, 1286, 2446, 996, 1569, 1570, 2452, 1006, 
       1535, 2584, 1536, 1016, 1158, 2458, 1026, 1167, 1038, 2488, 1177, 1051,
        2499, 1190, 1065}], 
      Line3DBox[{912, 1079, 2419, 913, 1092, 2429, 935, 1228, 2524, 1229, 
       1503, 887, 2528, 1594, 1236, 1238, 1204, 2512, 1247, 665, 1249, 2532, 
       666, 1262, 499, 1340, 2548, 539, 1325, 541, 2559, 1342, 548, 1330, 693,
        2574, 1288, 845, 1571, 2599, 1208, 1344, 786, 1537, 2585, 1211, 1464, 
       1454, 2570, 1455, 1027, 1541, 2588, 1542, 1039, 2489, 1178, 1052, 2500,
        1191, 1066}], Line3DBox[CompressedData["
1:eJwV0DsvBFEYBuBPFmtERVRIqNj4A34Gq3NJROHSUFhalx+gWNFQUohsQi0k
G3FtZiuXSCSiIRTUawvPFE/e77xn5pzJDMwuF5eaImKGci5ipDmilETsmG/M
L+ZbeccJKa+6U1ljtCXijTFS/YNu3FylyJP1hCyTa4+YzJ7XTctLpni3npN1
7+/KeX50i/KKBX6tP+zvmT/Z5093L7tafZ/scH5Dd2D+5pA8Nbo9028/NSes
cEyJTtb4cv62XM3OZJ0KG/TQsL8lH9mkl3MueM7+STa7/0y2ua/PfYnMU9UX
5CAF/VC2r7v23pEc5h+adjgG
       "]], 
      Line3DBox[{8, 1612, 23, 1626, 38, 293, 1904, 796, 53, 736, 2224, 304, 
       534, 68, 742, 2230, 318, 535, 83, 1669, 744, 333, 537, 98, 1686, 748, 
       349, 805, 2257, 113, 754, 367, 811, 2261, 128, 599, 383, 1960, 817, 
       143, 606, 396, 1974, 821, 158, 613, 2132, 407, 719, 173, 764, 2235, 
       414, 188, 1742, 879, 203, 1756, 218}], Line3DBox[CompressedData["
1:eJwVzy8sxHEcxvGPOTcu3fwJZnfmzyRNMtEumTFNwQQBY1NcE9gEgoDCRYds
IwkKG0lRTsHOdsXfgODmJbz3PJ/n83w/v/06phfHFmoiYhLbtRH9iYgdmklG
dKcicuYjc9bcjq+6iG+Myk/kD/wjWnRP6RnG7c7tJvhnlBvcp0/yQZp2Z4q/
0pk1J73dpXN4l+doo84Mf6tTb78vq6CAX/k1vcE8X3G/pHdgPkQRfSjhHk3e
L+mVdZbNL8ijGUPosm/1vbzOq86q7A1ryODO/WHaprOi86OzYf7AJqr2nXQE
6/YJXOBTPiDb4tPo+f8/2R6fxSV/THvxB/pENco=
       "]], 
      Line3DBox[{918, 1082, 2421, 919, 1589, 1094, 937, 1526, 2583, 1105, 948,
        1425, 2562, 1114, 958, 840, 2598, 1534, 778, 968, 2603, 863, 1583, 
       858, 978, 2577, 723, 1326, 470, 2551, 988, 551, 1332, 473, 2553, 998, 
       515, 1335, 2515, 707, 1008, 709, 1337, 2519, 1152, 1018, 898, 2612, 
       1603, 896, 1029, 1317, 2544, 1169, 1041, 2490, 1180, 1055, 2503, 1194, 
       1069}], Line3DBox[{920, 1083, 2422, 921, 1095, 2430, 938, 2461, 1106, 
       949, 1591, 2606, 1115, 959, 2465, 1533, 1121, 969, 2471, 1582, 1127, 
       979, 2477, 1278, 1133, 2441, 989, 1290, 1139, 2447, 999, 1301, 1145, 
       2453, 1009, 1310, 2540, 1153, 1019, 1602, 2610, 1160, 1030, 2483, 1170,
        1042, 2491, 1181, 1056, 2504, 1195, 1070}], 
      Line3DBox[{922, 1084, 2423, 923, 1096, 2431, 939, 1107, 2436, 950, 2462,
        1116, 960, 2466, 1122, 970, 2472, 1128, 980, 2478, 1134, 2442, 990, 
       1140, 2448, 1000, 1146, 2454, 1010, 1154, 2457, 1020, 2480, 1161, 1031,
        2484, 1171, 1043, 2492, 1182, 1057, 2505, 1196, 1071}], 
      Line3DBox[{924, 1085, 2424, 925, 1097, 2432, 940, 1108, 2437, 951, 1595,
        2608, 1596, 961, 2467, 1252, 1253, 971, 2473, 1266, 1267, 981, 2479, 
       1279, 1280, 2443, 991, 1291, 1292, 2449, 1001, 1572, 1573, 2455, 1011, 
       1538, 2586, 1539, 1021, 1162, 2459, 1032, 1172, 1044, 2493, 1183, 1058,
        2506, 1197, 1072}], 
      Line3DBox[{926, 1086, 2425, 927, 1098, 2433, 941, 1232, 2526, 1233, 
       1507, 890, 2529, 1597, 1240, 1242, 1205, 2513, 1254, 671, 1256, 2534, 
       672, 1268, 505, 1341, 2549, 542, 1327, 544, 2560, 1343, 553, 1333, 698,
        2575, 1294, 848, 1574, 2600, 1209, 1345, 789, 1540, 2587, 1212, 1465, 
       1457, 2571, 1458, 1033, 1546, 2590, 1547, 1045, 2494, 1184, 1059, 2507,
        1198, 1073}], Line3DBox[CompressedData["
1:eJwV0Lkvw2EYwPGnlKTiWKTGamLAnyOOtdqVdnasjs1gsIhjI2LAKCGODrRS
xyBNB4tEXCOLxec3fPJ93+d927dpvlgZK6ciosBOe8R1OqLKrvWNtjIRt1rj
iEdezI71ifGOiFcmaJg3zaasL5ikZV/QddJdEdP6ZlbSK4q828/on89v6Cw/
ZhWtUubXfkszvuNDt0lxR39nRF17nLXpHl/sJzMeyLqTd35v3cccB8wnZyzy
6f01XWCAJQ5ZJscq3+486wqDnHFO0/wk2fudp8m73st5r1e7uTTPJv8rm+4O
aV2H3R1hlH+w1Dea
       "]], Line3DBox[CompressedData["
1:eJwVzb8rxHEcx/G3OHKLH6mT7UQGBs5mUTIhg1IMFsdguWMwicUgrBQb9weY
KNsVxcImRUhKpHTRyUQehmev1/v1en2+3/R0fixXERFZtFdFFNCRiCjWRqT4
bZT4NlmS30Qf/ykr8r+VEat0VJbGj3yDfmAdZf2SvqU6YsR9pW+mKyhhGW82
izYpm9ZkxJCsHgt4xzwebfI2jfpb9w26sY8C9nBt8+r7c3bn7jN887v0BTuo
8f7CLitv8L/B///zs3QLCf0MPbGZktfZDLjv+En65PvPdALHNuPyI/4QTd4+
0Hsc6IZ1X/ylN2Xa61sZFHT9ui77HveaO2eTkZ2i0/YPOow3nA==
       "]]}, {}, {}}},
   VertexNormals->CompressedData["
1:eJxsXHc819/3N8sosyEpoowoLZJ1JaNhVKJSmUlDC5VVNCVKhKyMQhTZKuN9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    "]],
  Axes->True,
  BoxRatios->{1, 1, 0.4},
  ImageSize->{360., 285.71766121902573`},
  Method->{"RotationControl" -> "Globe"},
  PlotRange->{All, All, All},
  PlotRangePadding->{Automatic, Automatic, Automatic},
  ViewPoint->{-0.3228655520675471, -3.1866093521179444`, 1.091457132590488},
  ViewVertical->{0., 0., 1.}]], "Output",
 CellChangeTimes->{
  3.707568022011072*^9},ImageCache->GraphicsData["CompressedBitmap", "\<\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\
\>"],ExpressionUUID->"70b35bca-9e2c-46bf-8f8a-060037ef2b23"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"(", 
    RowBox[{"16", " ", 
     SuperscriptBox["a", "2"], " ", 
     SuperscriptBox["b", "2"], " ", 
     SuperscriptBox["c", "2"]}], ")"}], "/", 
   SuperscriptBox[
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "2"}], " ", 
       SuperscriptBox["a", "2"], " ", 
       SuperscriptBox["b", "2"]}], "+", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         SuperscriptBox["a", "2"], "+", 
         SuperscriptBox["b", "2"]}], ")"}], " ", 
       SuperscriptBox["c", "2"]}], "+", 
      RowBox[{"2", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", 
          SuperscriptBox["a", "2"]}], "+", 
         SuperscriptBox["b", "2"]}], ")"}], " ", 
       SuperscriptBox["c", "2"], " ", 
       RowBox[{"Cos", "[", 
        RowBox[{"2", " ", "u"}], "]"}], " ", 
       SuperscriptBox[
        RowBox[{"Cosh", "[", "v", "]"}], "2"]}], "+", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"2", " ", 
          SuperscriptBox["a", "2"], " ", 
          SuperscriptBox["b", "2"]}], "+", 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            SuperscriptBox["a", "2"], "+", 
            SuperscriptBox["b", "2"]}], ")"}], " ", 
          SuperscriptBox["c", "2"]}]}], ")"}], " ", 
       RowBox[{"Cosh", "[", 
        RowBox[{"2", " ", "v"}], "]"}]}]}], ")"}], "2"]}], "/.", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"a", "\[Rule]", "1"}], ",", 
    RowBox[{"b", "\[Rule]", "1"}], ",", 
    RowBox[{"c", "\[Rule]", "1"}]}], "}"}]}]], "Input",
 CellChangeTimes->{{3.707630239217065*^9, 
  3.707630263554384*^9}},ExpressionUUID->"51fcdcf2-be68-4c22-9418-\
1f6a73cb6618"],

Cell[BoxData[
 SuperscriptBox[
  RowBox[{"Sech", "[", 
   RowBox[{"2", " ", "v"}], "]"}], "2"]], "Output",
 CellChangeTimes->{
  3.707630272863533*^9},ExpressionUUID->"cda46188-6e4a-4bb7-9912-\
d51598f03449"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
},
WindowSize->{740, 535},
WindowMargins->{{6, Automatic}, {Automatic, 66}},
Magnification:>1.25 Inherited,
FrontEndVersion->"12.0 for Linux x86 (64-bit) (April 8, 2019)",
StyleDefinitions->"Default.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[579, 22, 160, 3, 125, "Title",ExpressionUUID->"facb1aa4-862e-467b-94f5-098711a93ba6"],
Cell[742, 27, 193, 4, 59, "Text",ExpressionUUID->"339ff77c-1b17-45e7-95a4-cdad4aef9a2f"],
Cell[938, 33, 250, 8, 45, "Text",ExpressionUUID->"86ae8b41-ccbd-4b75-b94d-adb18473354e"],
Cell[1191, 43, 1037, 18, 313, "Text",ExpressionUUID->"dea1b839-90da-4ec8-af4e-67fff5ea1a65"],
Cell[2231, 63, 62418, 1642, 3102, "Input",ExpressionUUID->"a2de38dc-5fbd-41dd-bd0b-9032f262b5fd"],
Cell[CellGroupData[{
Cell[64674, 1709, 153, 3, 85, "Section",ExpressionUUID->"247e319f-dbbf-47bc-a839-ba778860ff55"],
Cell[64830, 1714, 151, 3, 45, "Text",ExpressionUUID->"ee6fbf9e-5083-40bb-a8f6-344d039722d4"],
Cell[CellGroupData[{
Cell[65006, 1721, 236, 6, 39, "Input",ExpressionUUID->"4f47a352-3ecd-4f46-9f6b-b8f168c29eea"],
Cell[65245, 1729, 437, 13, 44, "Output",ExpressionUUID->"5076a385-816d-45bc-beae-c875b6d47c2b"]
}, Open  ]],
Cell[CellGroupData[{
Cell[65719, 1747, 327, 8, 39, "Input",ExpressionUUID->"8a8834a5-a9b0-40df-abbe-19e9e154f133"],
Cell[66049, 1757, 5985, 169, 437, "Output",ExpressionUUID->"7ec8ea6a-ae5b-47af-86b2-301df4b4cd14"]
}, Open  ]],
Cell[CellGroupData[{
Cell[72071, 1931, 193, 4, 41, "Input",ExpressionUUID->"3b08beef-70a3-4a18-b042-90ca63468837"],
Cell[72267, 1937, 247, 6, 63, "Output",ExpressionUUID->"5839a286-4cb3-4ca1-8501-895ce68eadd4"]
}, Open  ]],
Cell[72529, 1946, 232, 4, 45, "Text",ExpressionUUID->"e55e5eac-1893-48ed-955b-aada8cdc0384"],
Cell[72764, 1952, 158, 3, 45, "Text",ExpressionUUID->"808252ce-4067-4535-8750-8412d9c4d59e"],
Cell[CellGroupData[{
Cell[72947, 1959, 186, 4, 39, "Input",ExpressionUUID->"1c36da4b-67e8-420c-87ea-9a64dd8c4df0"],
Cell[73136, 1965, 386, 11, 44, "Output",ExpressionUUID->"166d12cb-5210-4947-b003-499eaa36914a"]
}, Open  ]],
Cell[73537, 1979, 396, 11, 39, "Input",ExpressionUUID->"1f182eeb-66f9-4af4-bc26-61eaa99cc547"],
Cell[CellGroupData[{
Cell[73958, 1994, 208, 5, 39, "Input",ExpressionUUID->"9f9f2fed-a412-4011-95a8-5c2188ed9c38"],
Cell[74169, 2001, 936, 30, 74, "Output",ExpressionUUID->"c4955467-5c8d-49b2-93ba-30bb1e26c9cc"]
}, Open  ]],
Cell[CellGroupData[{
Cell[75142, 2036, 247, 6, 39, "Input",ExpressionUUID->"35d99c3c-3337-45e1-bea6-6e110e948b16"],
Cell[75392, 2044, 22316, 646, 936, "Output",ExpressionUUID->"2093f016-b8f7-49a9-9ee6-ed89f04eef28"]
}, Open  ]],
Cell[CellGroupData[{
Cell[97745, 2695, 384, 9, 39, "Input",ExpressionUUID->"be483a63-36f5-421a-85ed-53ff52ae5c26"],
Cell[98132, 2706, 228, 5, 44, "Output",ExpressionUUID->"3d72f217-b81a-4903-a977-5409191bc2f9"]
}, Open  ]],
Cell[98375, 2714, 194, 3, 45, "Text",ExpressionUUID->"4dca132c-2b2c-4387-9bf1-5ca997af2d68"],
Cell[98572, 2719, 189, 4, 45, "Text",ExpressionUUID->"73404bc8-7cf7-4ac1-a95f-494e17a28896"],
Cell[98764, 2725, 184, 6, 75, "Text",ExpressionUUID->"83a32c95-9c8f-4cf5-b7cf-01df489de978"],
Cell[98951, 2733, 492, 13, 39, "Input",ExpressionUUID->"6d5c337e-7648-4965-a07b-95def1ba1015"],
Cell[CellGroupData[{
Cell[99468, 2750, 272, 6, 39, "Input",ExpressionUUID->"ee3680d1-f5b2-4900-b550-af3344d7fac6"],
Cell[99743, 2758, 36449, 952, 2593, "Output",ExpressionUUID->"1320a40b-b05e-4e7e-a013-6eddae96f6e3"]
}, Open  ]],
Cell[CellGroupData[{
Cell[136229, 3715, 245, 6, 39, "Input",ExpressionUUID->"d623b481-c26e-422c-8713-70f059abaf92"],
Cell[136477, 3723, 130, 3, 44, "Output",ExpressionUUID->"8272bad4-8f8e-49de-8d60-876704fbe465"]
}, Open  ]],
Cell[136622, 3729, 193, 6, 75, "Text",ExpressionUUID->"808c9967-c62a-4d30-b709-5fcd58626c7b"],
Cell[136818, 3737, 633, 20, 69, "Input",ExpressionUUID->"fdcfcb46-e83f-4a18-88bf-92255a2aebd9"],
Cell[CellGroupData[{
Cell[137476, 3761, 208, 5, 39, "Input",ExpressionUUID->"6ff47261-b161-4e8b-9edb-544ab45046d9"],
Cell[137687, 3768, 483, 16, 60, "Output",ExpressionUUID->"7460d1eb-028e-4cd1-8644-bdcb22b2c233"]
}, Open  ]],
Cell[CellGroupData[{
Cell[138207, 3789, 247, 6, 39, "Input",ExpressionUUID->"6cbfaa6e-f14f-45d4-8b6f-4ac0d0f3886c"],
Cell[138457, 3797, 5071, 162, 632, "Output",ExpressionUUID->"f65ae86b-47c0-4141-a330-837b73f2fa2f"]
}, Open  ]],
Cell[CellGroupData[{
Cell[143565, 3964, 317, 9, 39, "Input",ExpressionUUID->"454b7483-6b87-4e54-9358-0140203bb7de"],
Cell[143885, 3975, 132, 3, 44, "Output",ExpressionUUID->"a8cfd252-54ba-4218-b945-b43251753e20"]
}, Open  ]],
Cell[144032, 3981, 189, 6, 75, "Text",ExpressionUUID->"fd306329-504f-4a61-b044-cb31cf26760e"],
Cell[CellGroupData[{
Cell[144246, 3991, 251, 7, 39, "Input",ExpressionUUID->"72bc9c41-81e6-40ca-9e12-13717653be1c"],
Cell[144500, 4000, 440, 13, 44, "Output",ExpressionUUID->"96e9451c-b277-40a3-a160-d5a8c6f93cf8"]
}, Open  ]],
Cell[CellGroupData[{
Cell[144977, 4018, 318, 8, 39, "Input",ExpressionUUID->"6ada5c36-d66d-4880-a70e-3ec2d60f8d6a"],
Cell[145298, 4028, 6010, 165, 500, "Output",ExpressionUUID->"2b24126e-98b4-4deb-a2bd-435ca33f5fcd"]
}, Open  ]],
Cell[CellGroupData[{
Cell[151345, 4198, 193, 4, 39, "Input",ExpressionUUID->"7f565a74-40e8-479b-968d-3260a891df7e"],
Cell[151541, 4204, 1541, 48, 114, "Output",ExpressionUUID->"b8876d13-97f8-45d2-a1e0-827336bceae5"]
}, Open  ]],
Cell[CellGroupData[{
Cell[153119, 4257, 708, 20, 69, "Input",ExpressionUUID->"af18cf9a-e8d7-4608-86a9-cd5a859f0502"],
Cell[153830, 4279, 209834, 3446, 378, 137803, 2264, "CachedBoxData", "BoxData", "Output",ExpressionUUID->"70b35bca-9e2c-46bf-8f8a-060037ef2b23"]
}, Open  ]],
Cell[CellGroupData[{
Cell[363701, 7730, 1696, 53, 139, "Input",ExpressionUUID->"51fcdcf2-be68-4c22-9418-1f6a73cb6618"],
Cell[365400, 7785, 208, 6, 44, "Output",ExpressionUUID->"cda46188-6e4a-4bb7-9912-d51598f03449"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}
]
*)

